/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import java.io.Serializable;
import java.util.Comparator;
import java.util.NoSuchElementException;

import static jaggregate.Objects.*;
import static jaggregate.internal.ArgumentChecks.*;
import jaggregate.internal.Casting;
import jaggregate.internal.NonLocalReturnException;

/**
 * Implementation of the <dfn>collection</dfn> concept that should be common for most
 * concrete implementations.
 *
 * @param <E> a restriction on the types of elements that may be included in a collection
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: AbstractCollection.java,v 1.8 2008/10/03 19:01:23 pholser Exp $
 */
public abstract class AbstractCollection<E> implements Collection<E> {
    /**
     * Creates an empty collection.
     */
    protected AbstractCollection() {
        // For subclasses.
    }

    /**
     * {@inheritDoc}
     */
    public boolean allSatisfy( final UnaryCondition<? super E> discriminator ) {
        ensureNotNull( discriminator, DISCRIMINATOR );

        try {
            forEachDo( new UnaryFunctor<E, Void>() {
                public Void evaluate( E argument ) {
                    if ( !discriminator.matches( argument ) )
                        throw new NonLocalReturnException();
                    return null;
                }
            } );

            return true;
        }
        catch ( NonLocalReturnException ignored ) {
            return false;
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean anySatisfy( final UnaryCondition<? super E> discriminator ) {
        ensureNotNull( discriminator, DISCRIMINATOR );

        try {
            forEachDo( new UnaryFunctor<E, Void>() {
                public Void evaluate( E argument ) {
                    if ( discriminator.matches( argument ) )
                        throw new NonLocalReturnException();
                    return null;
                }
            } );

            return false;
        }
        catch ( NonLocalReturnException ignored ) {
            return true;
        }
    }

    /**
     * {@inheritDoc}
     */
    public Object[] toArray() {
        return toArray( Object.class );
    }

    /**
     * {@inheritDoc}
     */
    public Object[] toArray( Class<? super E> componentClass ) {
        return Collections.toArray( this, componentClass );
    }

    /**
     * {@inheritDoc}
     */
    public Bag<E> toBag() {
        return new Bag<E>( this );
    }

    /**
     * {@inheritDoc}
     */
    public IdentityBag<E> toIdentityBag() {
        return new IdentityBag<E>( this );
    }

    /**
     * {@inheritDoc}
     */
    public IdentitySet<E> toIdentitySet() {
        return new IdentitySet<E>( this );
    }

    /**
     * {@inheritDoc}
     */
    public OrderedCollection<E> toOrderedCollection() {
        return new OrderedCollection<E>( this );
    }

    /**
     * {@inheritDoc}
     */
    public Set<E> toSet() {
        return new Set<E>( this );
    }

    /**
     * {@inheritDoc}
     */
    public SortedCollection<E> toSortedCollection() {
        return new SortedCollection<E>( ByHashCode.INSTANCE, this );
    }

    /**
     * {@inheritDoc}
     */
    public SortedCollection<E> toSortedCollection( Comparator<? super E> comparator ) {
        return new SortedCollection<E>( comparator, this );
    }

    /**
     * {@inheritDoc}
     */
    public <R> Collection<R> collect(
        final UnaryFunctor<? super E, ? extends R> transformer ) {

        ensureNotNull( transformer, TRANSFORMER );

        final ExtensibleCollection<R> results = newEmptyExtensibleResultCollection();
        forEachDo( new UnaryFunctor<E, Void>() {
            public Void evaluate( E argument ) {
                results.add( transformer.evaluate( argument ) );
                return null;
            }
        } );

        return results;
    }

    /**
     * {@inheritDoc}
     */
    public E detect( final UnaryCondition<? super E> discriminator ) {
        ensureNotNull( discriminator, DISCRIMINATOR );

        final Object[] match = new Object[] { null };

        try {
            forEachDo( new UnaryFunctor<E, Void>() {
                public Void evaluate( E argument ) {
                    if ( discriminator.matches( argument ) ) {
                        match[ 0 ] = argument;
                        throw new NonLocalReturnException();
                    }

                    return null;
                }
            } );

            throw new NoSuchElementException();
        }
        catch ( NonLocalReturnException ignored ) {
            return Casting.<E>cast( match[ 0 ] );
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean includes( final E target ) {
        try {
            forEachDo( new UnaryFunctor<E, Void>() {
                public Void evaluate( E argument ) {
                    if ( target.equals( argument ) )
                        throw new NonLocalReturnException();
                    return null;
                }
            } );

            return false;
        }
        catch ( NonLocalReturnException ignored ) {
            return true;
        }
    }

    /**
     * {@inheritDoc}
     */
    public <R> R inject( R initialValue,
        final BinaryFunctor<? super R, ? super E, ? extends R> operation ) {

        ensureNotNull( operation, OPERATION );

        final Object[] result = new Object[] { initialValue };

        forEachDo( new UnaryFunctor<E, Void>() {
            public Void evaluate( E argument ) {
                result[ 0 ] =
                    operation.evaluate( Casting.<R>cast( result[ 0 ] ), argument );
                return null;
            }
        } );

        return Casting.<R>cast( result[ 0 ] );
    }

    /**
     * {@inheritDoc}
     */
    public boolean isEmpty() {
        return size() == 0;
    }

    /**
     * {@inheritDoc}
     */
    public int occurrencesOf( final E target ) {
        final int[] numberOfOccurrences = new int[] { 0 };

        forEachDo( new UnaryFunctor<E, Void>() {
            public Void evaluate( E argument ) {
                if ( target.equals( argument ) )
                    ++numberOfOccurrences[ 0 ];
                return null;
            }
        } );

        return numberOfOccurrences[ 0 ];
    }

    /**
     * {@inheritDoc}
     */
    public void rehash() {
        // does nothing by default.
    }

    /**
     * {@inheritDoc}
     */
    public Collection<E> reject( final UnaryCondition<? super E> discriminator ) {
        ensureNotNull( discriminator, DISCRIMINATOR );

        final ExtensibleCollection<E> rejects = newEmptyExtensibleCollection();

        forEachDo( new UnaryFunctor<E, Void>() {
            public Void evaluate( E argument ) {
                if ( !discriminator.matches( argument ) )
                    rejects.add( argument );
                return null;
            }
        } );

        return rejects;
    }

    /**
     * {@inheritDoc}
     */
    public Collection<E> select( final UnaryCondition<? super E> discriminator ) {
        ensureNotNull( discriminator, DISCRIMINATOR );

        final ExtensibleCollection<E> matches = newEmptyExtensibleCollection();

        forEachDo( new UnaryFunctor<E, Void>() {
            public Void evaluate( E argument ) {
                if ( discriminator.matches( argument ) )
                    matches.add( argument );
                return null;
            }
        } );

        return matches;
    }

    /**
     * Answers a new, empty extensible collection for use internally.
     *
     * @return a new, empty extensible collection
     */
    protected abstract ExtensibleCollection<E> newEmptyExtensibleCollection();

    /**
     * Answers a new, empty extensible collection for use internally.
     *
     * @param <R> type of the elements in the new collection
     * @return a new, empty extensible collection
     */
    protected abstract <R> ExtensibleCollection<R> newEmptyExtensibleResultCollection();

    private static class ByHashCode implements Comparator<Object>, Serializable {
        static final ByHashCode INSTANCE = new ByHashCode();
        private static final long serialVersionUID = -1L;

        public int compare( Object first, Object second ) {
            int firstHash = nullSafeHashCode( first );
            int secondHash = nullSafeHashCode( second );

            if ( firstHash == secondHash )
                return 0;

            if ( firstHash < secondHash )
                return -1;

            return 1;
        }
    }
}
