/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

/**
 * Provides protocol for accessing, adding, removing, and iterating over the
 * <dfn>elements</dfn> of an unordered collection whose elements are accessed using
 * an explicitly assigned external <dfn>key</dfn>.
 *
 * @param <K> a restriction on the types of the keys that may be contained in the
 * dictionary
 * @param <V> a restriction on the types of the values that may be contained in the
 * dictionary
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: AbstractDictionary.java,v 1.5 2008/05/08 03:49:37 pholser Exp $
 */
public interface AbstractDictionary<K, V> extends ExtensibleCollection<Pair<K, V>> {
    /**
     * {@inheritDoc}
     *
     * @return a bag of results
     */
    <R> AbstractBag<R> collect(
        UnaryFunctor<? super Pair<K, V>, ? extends R> transformer );

    /**
     * {@inheritDoc}
     *
     * @return a dictionary of rejects
     */
    AbstractDictionary<K, V> reject( UnaryCondition<? super Pair<K, V>> discriminator );

    /**
     * {@inheritDoc}
     *
     * @return a dictionary of matches
     */
    AbstractDictionary<K, V> select( UnaryCondition<? super Pair<K, V>> discriminator );

    /**
     * Answers the element associated with the given key in this dictionary.
     *
     * @param key the key to look up
     * @return the associated element, or {@code null} if there is no such element
     */
    V at( K key );

    /**
     * Answers a new dictionary whose keys are this dictionary's keys, and whose
     * corresponding elements are the results of evaluating the given transformer with
     * each corresponding element of this dictionary.  For each of this dictionary's
     * keys, a new element is obtained by evaluating {@code transformer} with the
     * corresponding element of this dictionary as a parameter.
     * <p/>
     * The elements are traversed in the order specified by {@link
     * #forEachDo(UnaryFunctor) forEachDo} for this dictionary.
     *
     * @param <R> return type of the transformer
     * @param transformer the transformer to evaluate
     * @return the results of evaluating {@code transformer} with the elements of this
     * dictionary
     * @throws NullPointerException if {@code transformer} is {@code null}
     */
    <R> AbstractDictionary<K, R> collectValues(
        UnaryFunctor<? super V, ? extends R> transformer );

    /**
     * Answers {@code true} if this dictionary contains an element stored at the given
     * key, else answers {@code false}.
     *
     * @param key the key to lookup
     * @return whether the key is in this dictionary
     */
    boolean includesKey( K key );

    /**
     * Answers a key such that the element associated with this key is equivalent to
     * the given value.
     * <p/>
     * Note that if there are multiple elements in this dictionary that are equivalent to
     * the given value, then the one whose key answered is arbitrary.
     *
     * @param value the value to look up
     * @return an associated key, or {@code null} if there is no such key
     */
    K keyAt( V value );

    /**
     * Answers a set of keys at which there is an element stored in this dictionary.
     * <p/>
     * The {@linkplain #size() size} of the result is equal to the {@linkplain #size()
     * size} of this dictionary.
     *
     * @return a set of the keys in this dictionary
     */
    AbstractSet<K> keys();

    /**
     * Iteratively evaluates the given operation with each of this dictionary's keys and
     * values.
     * <p/>
     * For each of this dictionary's elements, {@code operation} is evaluated with the
     * corresponding key as the first argument and the element as the second argument.
     * <p/>
     * The order in which the elements are visited is not specified.  Each key is visited
     * exactly once.
     *
     * @param operation the operation to perform
     * @throws NullPointerException if {@code operation} is {@code null}
     */
    void keysAndValuesDo( BinaryFunctor<? super K, ? super V, ?> operation );

    /**
     * Iteratively evaluates the given operation with each of this dictionary's keys
     * at which there are elements stored.
     * <p/>
     * For each of this dictionary's elements, {@code operation} is evaluated with the
     * corresponding key as the argument.
     * <p/>
     * The order in which the elements are visited is not specified.  Each key is visited
     * exactly once.
     *
     * @param <R> a constraint on the return type of the operation
     * @param operation the operation to perform
     * @throws NullPointerException if {@code operation} is {@code null}
     */
    <R> void keysDo( UnaryFunctor<? super K, ? extends R> operation );

    /**
     * Stores the elements of the given dictionary in this dictionary at the
     * corresponding key from the given dictionary.
     * <p/>
     * This message is equivalent to repeatedly performing {@link #putAt(Object,Object)
     * putAt} on this dictionary with each of the keys and elements from {@code newPairs}
     * in turn.  If a key in {@code newPairs} is <dfn>key-equivalent</dfn> to a key in
     * this dictionary, the associated element in {@code newPairs} replaces the element
     * in this dictionary.
     *
     * @param newPairs the pairs to add
     * @throws NullPointerException if {@code newPairs} is {@code null}
     */
    void putAll( AbstractDictionary<? extends K, ? extends V> newPairs );

    /**
     * Stores a new element at a given key in this dictionary.
     * <p/>
     * If lookup succeeds for {@code key}, then {@code newElement} replaces the element
     * previously stored at {@code key}.  Otherwise, {@code newElement} is stored at the
     * new {@code key}.  In either case, subsequent successful lookups for {@code key}
     * will answer {@code newElement}.
     *
     * @param key a key
     * @param newElement the new element to associate with {@code key}
     * @return the element previously associated with {@code key}, or {@code null} if
     * there is no such element
     */
    V putAt( K key, V newElement );

    /**
     * Answers a new dictionary which excludes the elements in this dictionary that cause
     * the given discriminator to answer {@code true}.
     * <p/>
     * For each of this dictionary's keys, {@code discriminator} is evaluated with the
     * corresponding element as the argument.  If the element causes {@code discriminator}
     * to answer {@code false}, the key is added to the answer with the element as its
     * corresponding value.
     *
     * @param discriminator the discriminator to evaluate
     * @return a dictionary of rejects
     * @throws NullPointerException if {@code discriminator} is {@code null}
     */
    AbstractDictionary<K, V> rejectValues( UnaryCondition<? super V> discriminator );

    /**
     * Removes any elements from this dictionary which are stored at any of the given
     * keys.
     * <p/>
     * This message has the same effect as repeatedly performing {@link
     * #removeKey(Object) removeKey} on this dictionary for each element in
     * {@code oldKeys}.
     *
     * @param oldKeys the keys to remove
     * @throws NullPointerException if {@code oldKeys} is {@code null}
     */
    void removeAllKeys( Collection<? extends K> oldKeys );

    /**
     * @see #removeAllKeys(Collection)
     * @param oldKeys the keys to remove
     * @throws NullPointerException if {@code restOfOldKeys} is {@code null}
     */
    void removeAllKeys( K[] oldKeys );

    /**
     * @see #removeAllKeys(Collection)
     * @param oldKey the first key to remove
     * @param restOfOldKeys the remainder of keys to remove
     * @throws NullPointerException if {@code restOfOldKeys} is {@code null}
     */
    void removeAllKeys( K oldKey, K... restOfOldKeys );

    /**
     * @see #removeAllKeys(Collection)
     * @param oldKeys the keys to remove
     * @throws NullPointerException if {@code oldKeys} is {@code null}
     */
    void removeAllKeys( Iterable<? extends K> oldKeys );

    /**
     * Removes the element stored at the given key in this dictionary.
     *
     * @param key the key to remove
     * @return the removed element, or {@code null} if there is no such element
     */
    V removeKey( K key );

    /**
     * Removes each element of this dictionary whose key causes the given discriminator
     * to answer {@code true}.
     *
     * @param discriminator the discriminator to evaluate
     * @return {@code true} if any removal occurred
     * @throws NullPointerException if {@code discriminator} is {@code null}
     */
    boolean removeIfKey( UnaryCondition<? super K> discriminator );

    /**
     * Removes each element of this dictionary which causes the given discriminator to
     * answer {@code true}.
     *
     * @param discriminator the discriminator to evaluate
     * @return {@code true} if any removal occurred
     * @throws NullPointerException if {@code discriminator} is {@code null}
     */
    boolean removeIfValue( UnaryCondition<? super V> discriminator );

    /**
     * Removes any elements from this dictionary which are not stored at any of the given
     * keys.
     * <p/>
     * Note that if the given collection is {@linkplain #isEmpty() empty}, this method
     * has the effect of "clearing" this dictionary.
     *
     * @param keepers the keys to retain
     * @return {@code true} if any elements were removed from this dictionary
     * @throws NullPointerException if {@code keepers} is {@code null}
     */
    boolean retainAllKeys( Collection<? extends K> keepers );

    /**
     * @see #retainAllKeys(Collection)
     * @param keepers the keys to retain
     * @return {@code true} if any elements were removed from this dictionary
     * @throws NullPointerException if {@code keepers} is {@code null}
     */
    boolean retainAllKeys( K[] keepers );

    /**
     * @see #retainAllKeys(Collection)
     * @param keeper the first key to retain
     * @param restOfKeepers the remainder of keys to retain
     * @return {@code true} if any elements were removed from this dictionary
     * @throws NullPointerException if {@code restOfKeepers} is {@code null}
     */
    boolean retainAllKeys( K keeper, K... restOfKeepers );

    /**
     * @see #retainAllKeys(Collection)
     * @param keepers the keys to retain
     * @return {@code true} if any elements were removed from this dictionary
     * @throws NullPointerException if {@code keepers} is {@code null}
     */
    boolean retainAllKeys( Iterable<? extends K> keepers );

    /**
     * Removes each element of this dictionary whose key causes the given discriminator
     * to answer {@code false}.
     *
     * @param discriminator the discriminator to evaluate
     * @return {@code true} if any removal occurred
     * @throws NullPointerException if {@code discriminator} is {@code null}
     */
    boolean retainIfKey( UnaryCondition<? super K> discriminator );

    /**
     * Removes each element of this dictionary which causes the given discriminator to
     * answer {@code false}.
     *
     * @param discriminator the discriminator to evaluate
     * @return {@code true} if any removal occurred
     * @throws NullPointerException if {@code discriminator} is {@code null}
     */
    boolean retainIfValue( UnaryCondition<? super V> discriminator );

   /**
     * Answers a new dictionary which contains the elements in this dictionary whose keys
     * cause the given discriminator to answer {@code true}.
     * <p/>
     * For each of this dictionary's keys, {@code discriminator} is evaluated with the
     * corresponding element as the argument.  If the element causes {@code discriminator}
     * to answer {@code true}, the key is added to the answer with the element as
     * its corresponding value.
     *
     * @param discriminator the discriminator to evaluate
     * @return a dictionary of matches
     * @throws NullPointerException if {@code discriminator} is {@code null}
     */
    AbstractDictionary<K, V> selectValues( UnaryCondition<? super V> discriminator );

    /**
     * Answers a collection of this dictionary's elements.
     *
     * @return a bag of the values in this dictionary
     */
    AbstractBag<V> values();
}
