/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import jaggregate.internal.Casting;
import jaggregate.internal.EquivalenceTester;
import jaggregate.internal.HashingContainer;
import jaggregate.internal.NonLocalReturnException;
import static jaggregate.internal.ArgumentChecks.*;
import static jaggregate.internal.Casting.*;

/**
 * Implementation of the <dfn>dictionary</dfn> concept that should be common for most
 * concrete implementations.
 *
 * @param <K> a restriction on the types of the keys that may be contained in the
 * dictionary
 * @param <V> a restriction on the types of the values that may be contained in the
 * dictionary
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: AbstractDictionaryImpl.java,v 1.10 2008/10/03 19:01:23 pholser Exp $
 */
public abstract class AbstractDictionaryImpl<K, V>
    extends AbstractExtensibleCollection<Pair<K, V>>
    implements AbstractDictionary<K, V> {

    private transient HashingContainer<K, V> hash;

    /**
     * For subclasses, so they can serialize themselves.
     */
    protected AbstractDictionaryImpl() {
        // nothing to do
    }

    /**
     * Creates an empty dictionary that uses the given strategy to determine key
     * equivalence.
     *
     * @param equivalenceTester a key equivalence strategy
     */
    protected AbstractDictionaryImpl( EquivalenceTester equivalenceTester ) {
        hash = new HashingContainer<K, V>( equivalenceTester );
    }

    /**
     * Creates a dictionary that contains the given associations and uses the given
     * strategy to determine key equivalence.
     *
     * @param associations the associations for the new dictionary to have
     * @param equivalenceTester a key equivalence strategy
     */
    protected AbstractDictionaryImpl(
        AbstractDictionary<? extends K, ? extends V> associations,
        EquivalenceTester equivalenceTester ) {

        this( equivalenceTester );
        putAll( associations );
    }

    /**
     * {@inheritDoc}
     */
    public void add( Pair<K, V> newElement ) {
        if ( newElement == null )
            throw new IllegalArgumentException( "cannot add a null pair" );

        putAt( newElement.key(), newElement.value() );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public <R> AbstractBag<R> collect(
        UnaryFunctor<? super Pair<K, V>, ? extends R> transformer ) {

        return (AbstractBag<R>) super.collect( transformer );
    }

    /**
     * {@inheritDoc}
     */
    public void forEachDo( UnaryFunctor<? super Pair<K, V>, ?> operation ) {
        ensureNotNull( operation, OPERATION );

        hash.forEachDo( operation );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean includes( Pair<K, V> target ) {
        if ( target == null )
            return false;

        K key = target.key();
        return includesKey( key ) && hash.areEqual( at( key ), target.value() );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int occurrencesOf( Pair<K, V> target ) {
        return includes( target ) ? 1 : 0;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void rehash() {
        hash.rehash();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AbstractDictionary<K, V> reject(
        UnaryCondition<? super Pair<K, V>> discriminator ) {

        return (AbstractDictionary<K, V>) super.reject( discriminator );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AbstractDictionary<K, V> select(
        UnaryCondition<? super Pair<K, V>> discriminator ) {

        return (AbstractDictionary<K, V>) super.select( discriminator );
    }

    /**
     * {@inheritDoc}
     */
    public int size() {
        return hash.size();
    }

    /**
     * {@inheritDoc}
     */
    public void putAll( AbstractDictionary<? extends K, ? extends V> newPairs ) {
        newPairs.keysAndValuesDo( new BinaryFunctor<K, V, Void>() {
            public Void evaluate( K first, V second ) {
                putAt( first, second );
                return null;
            }
        } );
    }

    /**
     * {@inheritDoc}
     */
    public V at( K key ) {
        return hash.get( key );
    }

    /**
     * {@inheritDoc}
     */
    public <R> AbstractDictionary<K, R> collectValues(
        final UnaryFunctor<? super V, ? extends R> transformer ) {

        ensureNotNull( transformer, TRANSFORMER );

        final AbstractDictionary<K, R> results = newEmptyDictionary();

        keysAndValuesDo( new BinaryFunctor<K, V, Void>() {
            public Void evaluate( K first, V second ) {
                results.putAt( first, transformer.evaluate( second ) );
                return null;
            }
        } );

        return results;
    }

    /**
     * {@inheritDoc}
     */
    public boolean includesKey( K key ) {
        return hash.hasKey( key );
    }

    /**
     * {@inheritDoc}
     */
    public K keyAt( final V value ) {
        final Object[] correspondingKey = new Object[] { null };

        try {
            keysAndValuesDo( new BinaryFunctor<K, V, Void>() {
                public Void evaluate( K first, V second ) {
                    if ( hash.areEqual( value, second ) ) {
                        correspondingKey[ 0 ] = first;
                        throw new NonLocalReturnException();
                    }

                    return null;
                }
            } );

            return null;
        }
        catch ( NonLocalReturnException ignored ) {
            return Casting.<K>cast( correspondingKey[ 0 ] );
        }
    }

    /**
     * {@inheritDoc}
     */
    public AbstractSet<K> keys() {
        final AbstractSet<K> keys = newEmptySet();

        keysDo( new UnaryFunctor<K, Void>() {
            public Void evaluate( K argument ) {
                keys.add( argument );
                return null;
            }
        } );

        return keys;
    }

    /**
     * {@inheritDoc}
     */
    public void keysAndValuesDo(
        final BinaryFunctor<? super K, ? super V, ?> operation ) {

        ensureNotNull( operation, OPERATION );

        forEachDo( new UnaryFunctor<Pair<K, V>, Void>() {
            public Void evaluate( Pair<K, V> argument ) {
                operation.evaluate( argument.key(), argument.value() );
                return null;
            }
        } );
    }

    /**
     * {@inheritDoc}
     */
    public <R> void keysDo( final UnaryFunctor<? super K, ? extends R> operation ) {
        ensureNotNull( operation, OPERATION );

        forEachDo( new UnaryFunctor<Pair<K, V>, Void>() {
            public Void evaluate( Pair<K, V> argument ) {
                operation.evaluate( argument.key() );
                return null;
            }
        } );
    }

    /**
     * {@inheritDoc}
     */
    public V putAt( K key, V newElement ) {
        return hash.put( key, newElement );
    }

    /**
     * {@inheritDoc}
     */
    public AbstractDictionary<K, V> rejectValues(
        final UnaryCondition<? super V> discriminator ) {

        ensureNotNull( discriminator, DISCRIMINATOR );

        final AbstractDictionary<K, V> rejects = newEmptyDictionary();

        keysAndValuesDo( new BinaryFunctor<K, V, Void>() {
            public Void evaluate( K first, V second ) {
                if ( !discriminator.matches( second ) )
                    rejects.putAt( first, second );

                return null;
            }
        } );

        return rejects;
    }

    /**
     * {@inheritDoc}
     */
    public boolean remove( Pair<K, V> oldElement ) {
        if ( includes( oldElement ) ) {
            removeKey( oldElement.key() );
            return true;
        }

        return false;
    }

    /**
     * {@inheritDoc}
     */
    public V removeKey( K key ) {
        return hash.remove( key );
    }

    /**
     * {@inheritDoc}
     */
    public void removeAllKeys( Collection<? extends K> oldKeys ) {
        oldKeys.forEachDo( new UnaryFunctor<K, Void>() {
            public Void evaluate( K argument ) {
                removeKey( argument );
                return null;
            }
        } );
    }

    /**
     * {@inheritDoc}
     */
    public void removeAllKeys( K[] oldKeys ) {
        for ( K each : oldKeys )
            removeKey( each );
    }

    /**
     * {@inheritDoc}
     */
    public void removeAllKeys( K oldKey, K... restOfOldKeys ) {
        removeKey( oldKey );
        removeAllKeys( restOfOldKeys );
    }

    /**
     * {@inheritDoc}
     */
    public void removeAllKeys( Iterable<? extends K> oldKeys ) {
        for ( K each : oldKeys )
            removeKey( each );
    }

    /**
     * {@inheritDoc}
     */
    public boolean removeIfKey( final UnaryCondition<? super K> discriminator ) {
        ensureNotNull( discriminator, DISCRIMINATOR );

        final boolean[] removalOccurred = new boolean[] { false };

        keysDo( new UnaryFunctor<K, Void>() {
            public Void evaluate( K argument ) {
                if ( discriminator.matches( argument ) ) {
                    removalOccurred[ 0 ] = true;
                    removeKey( argument );
                }

                return null;
            }
        } );

        return removalOccurred[ 0 ];
    }

    /**
     * {@inheritDoc}
     */
    public boolean removeIfValue( final UnaryCondition<? super V> discriminator ) {
        ensureNotNull( discriminator, DISCRIMINATOR );

        final boolean[] removalOccurred = new boolean[] { false };

        keysAndValuesDo( new BinaryFunctor<K, V, Void>() {
            public Void evaluate( K first, V second ) {
                if ( discriminator.matches( second ) ) {
                    removalOccurred[ 0 ] = true;
                    removeKey( first );
                }

                return null;
            }
        } );

        return removalOccurred[ 0 ];
    }

    private boolean retainAllKeysFrom( final Collection<Object> keepers ) {
        return removeIfKey( new UnaryCondition<K>() {
            public boolean matches( K target ) {
                return !keepers.includes( target );
            }
        } );
    }

    /**
     * {@inheritDoc}
     */
    public boolean retainAllKeys( Collection<? extends K> keepers ) {
        AbstractSet<Object> setOfKeepers = newEmptySet();
        setOfKeepers.addAll( keepers );

        return retainAllKeysFrom( setOfKeepers );
    }

    /**
     * {@inheritDoc}
     */
    public boolean retainAllKeys( K[] keepers ) {
        AbstractSet<Object> setOfKeepers = newEmptySet();
        setOfKeepers.addAll( keepers );

        return retainAllKeysFrom( setOfKeepers );
    }

    /**
     * {@inheritDoc}
     */
    public boolean retainAllKeys( K keeper, K... restOfKeepers ) {
        AbstractSet<Object> setOfKeepers = newEmptySet();
        setOfKeepers.add( keeper );
        setOfKeepers.addAll( restOfKeepers );

        return retainAllKeysFrom( setOfKeepers );
    }

    /**
     * {@inheritDoc}
     */
    public boolean retainAllKeys( Iterable<? extends K> keepers ) {
        AbstractSet<Object> setOfKeepers = newEmptySet();
        setOfKeepers.addAll( keepers );

        return retainAllKeysFrom( setOfKeepers );
    }

    /**
     * {@inheritDoc}
     */
    public boolean retainIfKey( final UnaryCondition<? super K> discriminator ) {
        ensureNotNull( discriminator, DISCRIMINATOR );

        final boolean[] removalOccurred = new boolean[] { false };

        keysDo( new UnaryFunctor<K, Void>() {
            public Void evaluate( K argument ) {
                if ( !discriminator.matches( argument ) ) {
                    removalOccurred[ 0 ] = true;
                    removeKey( argument );
                }

                return null;
            }
        } );

        return removalOccurred[ 0 ];
    }

    /**
     * {@inheritDoc}
     */
    public boolean retainIfValue( final UnaryCondition<? super V> discriminator ) {
        ensureNotNull( discriminator, DISCRIMINATOR );

        final boolean[] removalOccurred = new boolean[] { false };

        keysAndValuesDo( new BinaryFunctor<K, V, Void>() {
            public Void evaluate( K first, V second ) {
                if ( !discriminator.matches( second ) ) {
                    removalOccurred[ 0 ] = true;
                    removeKey( first );
                }

                return null;
            }
        } );

        return removalOccurred[ 0 ];
    }

    /**
     * {@inheritDoc}
     */
    public AbstractDictionary<K, V> selectValues(
        final UnaryCondition<? super V> discriminator ) {

        ensureNotNull( discriminator, DISCRIMINATOR );

        final AbstractDictionary<K, V> matches = newEmptyDictionary();

        keysAndValuesDo( new BinaryFunctor<K, V, Void>() {
            public Void evaluate( K first, V second ) {
                if ( discriminator.matches( second ) )
                    matches.putAt( first, second );

                return null;
            }
        } );

        return matches;
    }

    /**
     * {@inheritDoc}
     */
    public AbstractBag<V> values() {
        final AbstractBag<V> values = newEmptyExtensibleResultCollection();

        keysAndValuesDo( new BinaryFunctor<K, V, Void>() {
            public Void evaluate( K first, V second ) {
                values.add( second );
                return null;
            }
        } );

        return values;
    }

    /**
     * {@inheritDoc}
     * <p/>
     * An object is considered equivalent to this dictionary if the object is assignable
     * to this dictionary's class and has equivalent contents, as defined by this
     * dictionary's notion of equivalence.
     */
    @Override
    public boolean equals( Object that ) {
        if ( this == that )
            return true;

        if ( !getImplementationClass().isInstance( that ) )
            return false;

        final AbstractDictionary<Object, Object> other = cast( that );

        if ( size() != other.size() )
            return false;

        try {
            keysAndValuesDo( new DictionaryEqualityTester( other ) );
            return true;
        }
        catch ( NonLocalReturnException ignored ) {
            return false;
        }
    }

    /**
     * Subclasses implement this method to answer the root of its true runtime class,
     * for purposes of determining {@linkplain #equals(Object) equivalence}.
     *
     * @return this dictionary's implementation class
     */
    protected abstract Class<? extends AbstractDictionary> getImplementationClass();

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return inject( 0, new BinaryFunctor<Integer, Pair<K, V>, Integer>() {
            public Integer evaluate( Integer first, Pair<K, V> second ) {
                return first + second.hashCode();
            }
        } );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        final StringBuilder buffer = new StringBuilder( "{" );

        if ( !isEmpty() ) {
            hash.forEachDo( new UnaryFunctor<Pair<K, V>, Void>() {
                public Void evaluate( Pair<K, V> argument ) {
                    buffer.append( '[' );
                    buffer.append( asStringOrThis( argument.key() ) );
                    buffer.append( '=' );
                    buffer.append( asStringOrThis( argument.value() ) );
                    buffer.append( "]," );
                    return null;
                }
            } );

            buffer.deleteCharAt( buffer.length() - 1 );
        }

        buffer.append( '}' );

        return buffer.toString();
    }

    /**
     * Answers a new, empty dictionary for internal use.
     *
     * @param <T> a restriction on the types of keys allowed in the returned
     * dictionary
     * @param <U> a restriction on the types of values allowed in the returned
     * dictionary
     * @return a new, empty dictionary
     */
    protected abstract <T, U> AbstractDictionaryImpl<T, U> newEmptyDictionary();

    /**
     * {@inheritDoc}
     */
    @Override
    protected final AbstractDictionaryImpl<K, V> newEmptyExtensibleCollection() {
        return newEmptyDictionary();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected abstract <R> AbstractBag<R> newEmptyExtensibleResultCollection();

    /**
     * Tells how the given key or value will show up in this dictionary's string
     * representation.
     *
     * @param keyOrValue an element of this collection
     * @return a string representation of {@code keyOrValue} to be used in this bag's
     * string representation
     */
    protected abstract String asString( Object keyOrValue );

    private String asStringOrThis( Object keyOrValue ) {
        return keyOrValue == this ? "this" : asString( keyOrValue );
    }

    private int capacity() {
        return hash.capacity();
    }

    private void initialize( int capacity, EquivalenceTester tester ) {
        hash = new HashingContainer<K, V>( capacity, tester );
    }

    /**
     * Serialization helper method.
     *
     * @param output stream to write object state to
     * @throws IOException if an I/O problem occurs during serialization
     */
    protected final void serializeTo( final ObjectOutputStream output )
        throws IOException {

        output.defaultWriteObject();
        output.writeInt( capacity() );
        output.writeInt( size() );

        hash.forEachDo( new UnaryFunctor<Pair<K, V>, Void>() {
            public Void evaluate( Pair<K, V> argument ) {
                try {
                    output.writeObject( argument.key() );
                    output.writeObject( argument.value() );
                    return null;
                }
                catch ( IOException ex ) {
                    throw new RuntimeException( ex );
                }
            }
        } );
    }

    /**
     * De-serialization helper method.
     *
     * @param input stream to read object state from
     * @param tester equivalence tester to give the reconstituted dictionary
     * @throws IOException if an I/O problem occurs during de-serialization
     * @throws ClassNotFoundException if a class that is necessary for de-serialization
     * cannot be found
     */
    protected final void deserializeFrom( ObjectInputStream input,
        EquivalenceTester tester ) throws IOException, ClassNotFoundException {

        input.defaultReadObject();

        int capacity = input.readInt();
        initialize( capacity, tester );

        int size = input.readInt();
        for ( int i = 0; i < size; ++i ) {
            K key = Casting.<K>cast( input.readObject() );
            V value = Casting.<V>cast( input.readObject() );

            putAt( key, value );
        }
    }

    private class DictionaryEqualityTester
        implements BinaryFunctor<Object, Object, Void> {

        private final AbstractDictionary<Object, Object> other;

        public DictionaryEqualityTester( AbstractDictionary<Object, Object> other ) {
            this.other = other;
        }

        public Void evaluate( Object first, Object second ) {
            Object valueAtKey = other.at( first );

            if ( second == null ) {
                if ( !( valueAtKey == null && other.includesKey( first ) ) )
                    throw new NonLocalReturnException();
            }
            else {
                if ( !hash.areEqual( second, valueAtKey ) )
                    throw new NonLocalReturnException();
            }

            return null;
        }
    }
}
