/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import static jaggregate.internal.Casting.*;

/**
 * Represents an unordered, variable-sized collection whose <dfn>elements</dfn> can be
 * added or removed, but cannot be individually accessed by external <dfn>keys</dfn>.
 * <p/>
 * A set is similar to a {@linkplain Bag bag} but cannot contain duplicate elements.
 * Elements are duplicates if they are <dfn>equivalent</dfn>.
 *
 * @param <E> a restriction on the types of elements that may be included in the set
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: AbstractSet.java,v 1.8 2008/10/03 19:01:23 pholser Exp $
 */
public abstract class AbstractSet<E> extends AbstractExtensibleCollection<E> {
    private transient AbstractDictionary<E, Object> storage;

    /**
     * Creates an empty set.
     */
    protected AbstractSet() {
        initialize();
    }

    /**
     * Creates a set containing the elements in the given collection.
     *
     * @param elements elements to add to the new set
     * @throws NullPointerException if {@code elements} is {@code null}
     */
    protected AbstractSet( Collection<? extends E> elements ) {
        this();
        addAll( elements );
    }

    /**
     * Creates a set containing the elements in the given array.
     *
     * @param elements elements to add to the new set
     * @throws NullPointerException if {@code elements} is {@code null}
     */
    protected AbstractSet( E... elements ) {
        this();
        addAll( elements );
    }

    /**
     * Creates a set containing the elements in the given iterable object.
     *
     * @param elements elements to add to the new set
     * @throws NullPointerException if {@code elements} is {@code null}
     */
    protected AbstractSet( Iterable<? extends E> elements ) {
        this();
        addAll( elements );
    }

    /**
     * {@inheritDoc}
     *
     * @return a set of results
     */
    @Override
    public <R> AbstractSet<R> collect(
        UnaryFunctor<? super E, ? extends R> transformer ) {

        return (AbstractSet<R>) super.collect( transformer );
    }

    /**
     * {@inheritDoc}
     */
    public void forEachDo( UnaryFunctor<? super E, ?> operation ) {
        storage.keysDo( operation );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean includes( E target ) {
        return storage.includesKey( target );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int occurrencesOf( E target ) {
        return includes( target ) ? 1 : 0;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void rehash() {
        storage.rehash();
    }

    /**
     * {@inheritDoc}
     *
     * @return a set of rejects
     */
    @Override
    public AbstractSet<E> reject( UnaryCondition<? super E> discriminator ) {
        return (AbstractSet<E>) super.reject( discriminator );
    }

    /**
     * {@inheritDoc}
     *
     * @return a set of selections
     */
    @Override
    public AbstractSet<E> select( UnaryCondition<? super E> discriminator ) {
        return (AbstractSet<E>) super.select( discriminator );
    }

    /**
     * {@inheritDoc}
     */
    public int size() {
        return storage.size();
    }

    /**
     * {@inheritDoc}
     */
    public void add( E newElement ) {
        storage.putAt( newElement, null );
    }

    /**
     * {@inheritDoc}
     */
    public boolean remove( E oldElement ) {
        if ( includes( oldElement ) ) {
            storage.removeKey( oldElement );
            return true;
        }

        return false;
    }

    /**
     * {@inheritDoc}
     * <p/>
     * An object is considered equivalent to this set if the object is assignable to
     * this set's class and has equivalent contents, as defined by the this set's notion
     * of equivalence.
     */
    @Override
    public boolean equals( Object that ) {
        if ( this == that )
            return true;

        if ( !getImplementationClass().isInstance( that ) )
            return false;

        AbstractSet<?> other = (AbstractSet<?>) that;
        return storage.equals( other.storage );
    }

    /**
     * Subclasses implement this method to answer the root of its true runtime class,
     * for purposes of determining {@linkplain #equals(Object) equivalence}.
     *
     * @return this set's implementation class
     */
    protected abstract Class<? extends AbstractSet> getImplementationClass();

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return storage.hashCode();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        final StringBuilder buffer = new StringBuilder( "[" );

        if ( !isEmpty() ) {
            storage.keysDo( new UnaryFunctor<E, Void>() {
                public Void evaluate( E argument ) {
                    buffer.append( asString( argument ) ).append( ',' );
                    return null;
                }
            } );

            buffer.deleteCharAt( buffer.length() - 1 );
        }

        buffer.append( ']' );

        return buffer.toString();
    }

    /**
     * Answers a new, empty storage for the set.
     *
     * @return a new empty dictionary
     */
    protected abstract AbstractDictionary<E, Object> newEmptyStorage();

    /**
     * Tells how the given element will show up in this set's string representation.
     *
     * @param anElement an element of this collection
     * @return a string representation of {@code anElement} to be used in this set's
     * string representation
     */
    protected abstract String asString( E anElement );

    /**
     * Serialization helper method.
     *
     * @param output stream to write object state to
     * @throws IOException if an I/O problem occurs during serialization
     */
    protected final void serializeTo( ObjectOutputStream output ) throws IOException {
        output.defaultWriteObject();
        output.writeObject( storage );
    }

    /**
     * De-serialization helper method.
     *
     * @param input stream to read object state from
     * @throws IOException if an I/O problem occurs during de-serialization
     * @throws ClassNotFoundException if a class that is necessary for de-serialization
     * cannot be found
     */
    protected final void deserializeFrom( ObjectInputStream input )
        throws IOException, ClassNotFoundException {

        input.defaultReadObject();

        AbstractDictionary<E, Object> backingStore = cast( input.readObject() );
        initialize( backingStore );
    }

    private void initialize() {
        initialize( newEmptyStorage() );
    }

    private void initialize( AbstractDictionary<E, Object> backingStore ) {
        storage = backingStore;
    }
}
