/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

/**
 * A functor that accepts two arguments.
 *
 * @param <A1> a constraint on the allowable types for the functor's first argument
 * @param <A2> a constraint on the allowable types for the functor's second argument
 * @param <R> a constraint on the allowable types for the functor's return value
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: BinaryFunction.java,v 1.5 2008/10/03 19:01:23 pholser Exp $
 */
public abstract class BinaryFunction<A1, A2, R> implements BinaryFunctor<A1, A2, R> {
    /**
     * Creates a new functor.
     */
    protected BinaryFunction() {
        // For subclasses.
    }

    /**
     * Answers a functor of one argument that evaluates in the same manner as this
     * functor would if evaluated using a fixed first argument and another argument.
     *
     * @param boundArgument the fixed argument used as the first argument to this
     * functor's {@link #evaluate(Object,Object) evaluate} method
     * @return a functor that accepts one argument, of the same type as this functor's
     * second argument type, that behaves as this functor would if always evaluated
     * using {@code boundArgument} as the first argument
     */
    public final UnaryFunction<A2, R> bindFirst( A1 boundArgument ) {
        return new FirstBoundBinaryFunction<A1, A2, R>( this, boundArgument );
    }

    /**
     * Answers a functor of one argument that evaluates in the same manner as this
     * functor would if evaluated using a fixed second argument and another argument.
     *
     * @param boundArgument the fixed argument used as the second argument to this
     * functor's {@link #evaluate(Object,Object) evaluate} method
     * @return a functor that accepts one argument, of the same type as this functor's
     * first argument type, that behaves as this functor would if always evaluated
     * using {@code boundArgument} as the second argument
     */
    public final UnaryFunction<A1, R> bindSecond( A2 boundArgument ) {
        return new SecondBoundBinaryFunction<A1, A2, R>( this, boundArgument );
    }

    private static class FirstBoundBinaryFunction<A1, A2, R>
        extends UnaryFunction<A2, R> {

        private final BinaryFunctor<A1, A2, R> wrapped;
        private final A1 boundArgument;

        FirstBoundBinaryFunction( BinaryFunctor<A1, A2, R> wrapped, A1 boundArgument ) {
            this.wrapped = wrapped;
            this.boundArgument = boundArgument;
        }

        public R evaluate( A2 argument ) {
            return wrapped.evaluate( boundArgument, argument );
        }
    }

    private static class SecondBoundBinaryFunction<A1, A2, R>
        extends UnaryFunction<A1, R> {

        private final BinaryFunctor<A1, A2, R> wrapped;
        private final A2 boundArgument;

        SecondBoundBinaryFunction( BinaryFunctor<A1, A2, R> wrapped, A2 boundArgument ) {
            this.wrapped = wrapped;
            this.boundArgument = boundArgument;
        }

        public R evaluate( A1 argument ) {
            return wrapped.evaluate( argument, boundArgument );
        }
    }
}
