/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import static java.lang.String.*;

import static jaggregate.internal.ArgumentChecks.*;

/**
 * A predicate that accepts two arguments.
 *
 * @param <A1> a constraint on the allowable types for the predicate's first argument
 * @param <A2> a constraint on the allowable types for the predicate's second argument
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: BinaryPredicate.java,v 1.6 2008/10/03 19:01:23 pholser Exp $
 */
public abstract class BinaryPredicate<A1, A2> implements BinaryCondition<A1, A2> {
    /**
     * Creates a new predicate.
     */
    protected BinaryPredicate() {
        // For subclasses.
    }

    /**
     * Answers a predicate that represents the logical <dfn>inverse</dfn> of this
     * predicate; wherever this predicate's {@link #matches(Object,Object) matches}
     * method would answer {@code true}, the inverse answers {@code false}; and vice
     * versa.
     *
     * @return the inverse of this predicate
     */
    public final BinaryPredicate<A1, A2> not() {
        return new InvertedBinaryPredicate<A1, A2>( this );
    }

    /**
     * Answers a predicate that represents the logical <dfn>inverse</dfn> of the given
     * predicate; wherever the given predicate's {@link #matches(Object,Object) matches}
     * method would answer {@code true}, the inverse answers {@code false}; and vice
     * versa.
     *
     * @param <T> constraint on the first type accepted by the predicate
     * @param <U> constraint on the second type accepted by the predicate
     * @param predicate the predicate to invert
     * @return the inverse of {@code predicate}
     * @throws NullPointerException if {@code predicate} is {@code null}
     */
    public static <T, U> BinaryPredicate<T, U> not( BinaryPredicate<T, U> predicate ) {
        return new InvertedBinaryPredicate<T, U>( predicate );
    }

    /**
     * Answers a predicate that represents the logical <dfn>disjunction</dfn> of this
     * predicate and another predicate.  The disjunction's {@link #matches(Object,Object)
     * matches} method answers {@code true} if either this predicate or the other
     * predicate would answer {@code true}.
     * <p/>
     * When the disjunction is evaluated, this predicate is always evaluated first.
     * The other predicate may not be evaluated; so do not depend on its evaluation for
     * side effects.
     *
     * @param other the "right-hand" operand of the disjunction
     * @return the logical disjunction of this predicate and {@code other}
     */
    public final BinaryPredicate<A1, A2> or(
        BinaryPredicate<? super A1, ? super A2> other ) {

        return new DisjunctiveBinaryPredicate<A1, A2>( this, other );
    }

    /**
     * Answers a predicate that represents the logical <dfn>conjunction</dfn> of this
     * predicate and another predicate.  The conjunction's {@link #matches(Object,Object)
     * matches} method answers {@code true} if both this predicate and the
     * other predicate would answer {@code true}.
     * <p/>
     * When the conjunction is evaluated, this predicate is always evaluated first.
     * The other predicate may not be evaluated; so do not depend on its evaluation for
     * side effects.
     *
     * @param other the "right-hand" operand of the conjunction
     * @return the logical conjunction of this predicate and {@code other}
     */
    public final BinaryPredicate<A1, A2> and(
        BinaryPredicate<? super A1, ? super A2> other ) {

        return new ConjunctiveBinaryPredicate<A1, A2>( this, other );
    }

    /**
     * Answers a predicate that represents the logical <dfn>exclusive disjunction</dfn>
     * of this predicate and another predicate.  The disjunction's
     * {@link #matches(Object,Object) matches} method answers {@code true} if either
     * this predicate or the other predicate, but not both, would answer {@code true}.
     * <p/>
     * When the disjunction is evaluated, both this predicate and the other predicate
     * will be evaluated.
     *
     * @param other the "right-hand" operand of the disjunction
     * @return the logical exclusive disjunction of this predicate and {@code other}
     */
    public final BinaryPredicate<A1, A2> xor(
        BinaryPredicate<? super A1, ? super A2> other ) {

        return new ExclusivelyDisjunctiveBinaryPredicate<A1, A2>( this, other );
    }

    /**
     * Answers a predicate of one argument that matches in the same manner as this
     * predicate would if evaluated using a fixed first argument and another argument.
     *
     * @param boundArgument the fixed argument used as the first argument to this
     * predicate's {@link #matches(Object,Object) matches} method
     * @return a predicate that accepts one argument, of the same type as this
     * predicate's second argument type, that behaves as this predicate would if always
     * evaluated using {@code boundArgument} as the first argument
     */
    public final UnaryPredicate<A2> bindFirst( A1 boundArgument ) {
        return new FirstBoundBinaryPredicate<A1, A2>( this, boundArgument );
    }

    /**
     * Answers a predicate of one argument that matches in the same manner as this
     * predicate would if evaluated using a fixed second argument and another argument.
     *
     * @param boundArgument the fixed argument used as the second argument to this
     * predicate's {@link #matches(Object,Object) matches} method
     * @return a predicate that accepts one argument, of the same type as this
     * predicate's first argument type, that behaves as this predicate would if always
     * evaluated using {@code boundArgument} as the second argument
     */
    public final UnaryPredicate<A1> bindSecond( A2 boundArgument ) {
        return new SecondBoundBinaryPredicate<A1, A2>( this, boundArgument );
    }

    /**
     * Gives a string template that describes the kinds of objects which match this
     * predicate when it has its first argument bound via {@link #bindFirst(Object)}.
     * <p/>
     * The string should have at least one format specification as described by
     * {@link String#format(String, Object...)}; that specification should expect to
     * get filled in with this predicate's first bound argument.
     * <p/>
     * If not overridden, this method answers the same as {@link #toString()}.
     *
     * @see #describe()
     * @return a descriptive string template
     */
    protected String getFirstBoundDescription() {
        return toString();
    }

    /**
     * Gives a string template that describes the kinds of objects which match this
     * predicate when it has its second argument bound via {@link #bindSecond(Object)}.
     * <p/>
     * The string should have at least one format specification as described by
     * {@link String#format(String, Object...)}; that specification should expect to
     * get filled in with this predicate's second bound argument.
     * <p/>
     * If not overridden, this method answers the same as
     * {@link #getFirstBoundDescription()}.
     *
     * @see #describe()
     * @return a descriptive string template
     */
    protected String getSecondBoundDescription() {
        return getFirstBoundDescription();
    }

    /**
     * Gives a string that describes the kinds of objects which match this predicate.
     * <p/>
     * If not overridden, this method answers the same as {@link #toString()}.
     *
     * @return a description of objects that meet this predicate's criteria
     */
    public String describe() {
        return toString();
    }

    private static class InvertedBinaryPredicate<A1, A2>
        extends BinaryPredicate<A1, A2> {

        private final BinaryPredicate<A1, A2> wrapped;

        InvertedBinaryPredicate( BinaryPredicate<A1, A2> wrapped ) {
            ensureNotNull( wrapped, DISCRIMINATOR );
            this.wrapped = wrapped;
        }

        public boolean matches( A1 first, A2 second ) {
            return !wrapped.matches( first, second );
        }

        @Override
        public String describe() {
            return format( "something other than <%1$s>", wrapped.describe() );
        }
    }

    private abstract static class CompositeBinaryPredicate<A1, A2>
        extends BinaryPredicate<A1, A2> {

        protected final BinaryPredicate<A1, A2> first;
        protected final BinaryPredicate<? super A1, ? super A2> second;

        protected CompositeBinaryPredicate( BinaryPredicate<A1, A2> first,
            BinaryPredicate<? super A1, ? super A2> second ) {

            ensureNotNull( second, "'right-hand' operand" );

            this.first = first;
            this.second = second;
        }
    }

    private static class DisjunctiveBinaryPredicate<A1, A2>
        extends CompositeBinaryPredicate<A1, A2> {

        DisjunctiveBinaryPredicate( BinaryPredicate<A1, A2> first,
            BinaryPredicate<? super A1, ? super A2> second ) {

            super( first, second );
        }

        public boolean matches( A1 first, A2 second ) {
            return this.first.matches( first, second )
                || this.second.matches( first, second );
        }

        @Override
        public String describe() {
            return format( "<%1$s> or <%2$s>", first.describe(), second.describe() );
        }
    }

    private static class ConjunctiveBinaryPredicate<A1, A2>
        extends CompositeBinaryPredicate<A1, A2> {

        ConjunctiveBinaryPredicate( BinaryPredicate<A1, A2> first,
            BinaryPredicate<? super A1, ? super A2> second ) {

            super( first, second );
        }

        public boolean matches( A1 first, A2 second ) {
            return this.first.matches( first, second )
                && this.second.matches( first, second );
        }

        @Override
        public String describe() {
            return format( "<%1$s> and <%2$s>", first.describe(), second.describe() );
        }
    }

    private static class ExclusivelyDisjunctiveBinaryPredicate<A1, A2>
        extends CompositeBinaryPredicate<A1, A2> {

        ExclusivelyDisjunctiveBinaryPredicate( BinaryPredicate<A1, A2> first,
            BinaryPredicate<? super A1, ? super A2> second ) {

            super( first, second );
        }

        public boolean matches( A1 first, A2 second ) {
            return this.first.matches( first, second )
                ^ this.second.matches( first, second );
        }

        @Override
        public String describe() {
            return format( "either <%1$s> or <%2$s> but not both", first.describe(),
                second.describe() );
        }
    }

    private static class FirstBoundBinaryPredicate<A1, A2> extends UnaryPredicate<A2> {
        private final BinaryPredicate<A1, A2> wrapped;
        private final A1 boundArgument;

        FirstBoundBinaryPredicate( BinaryPredicate<A1, A2> wrapped, A1 boundArgument ) {
            this.wrapped = wrapped;
            this.boundArgument = boundArgument;
        }

        public boolean matches( A2 target ) {
            return wrapped.matches( boundArgument, target );
        }

        @Override
        public String describe() {
            return format( wrapped.getFirstBoundDescription(), boundArgument );
        }
    }

    private static class SecondBoundBinaryPredicate<A1, A2> extends UnaryPredicate<A1> {
        private final BinaryPredicate<A1, A2> wrapped;
        private final A2 boundArgument;

        SecondBoundBinaryPredicate( BinaryPredicate<A1, A2> wrapped, A2 boundArgument ) {
            this.wrapped = wrapped;
            this.boundArgument = boundArgument;
        }

        public boolean matches( A1 target ) {
            return wrapped.matches( target, boundArgument );
        }

        @Override
        public String describe() {
            return format( wrapped.getSecondBoundDescription(), boundArgument );
        }
    }
}
