/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import java.lang.reflect.Array;

import static jaggregate.Dictionary.*;
import static jaggregate.IdentityDictionary.*;
import static jaggregate.internal.ArgumentChecks.*;
import static jaggregate.internal.Casting.*;

/**
 * Methods that should be on {@link Collection}, but aren't for reasons of type safety
 * or generics restrictions.
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: Collections.java,v 1.9 2008/10/03 19:01:23 pholser Exp $
 */
public class Collections {
    /**
     * Discourages instantiation.
     *
     * @throws UnsupportedOperationException always
     */
    protected Collections() {
        throw new UnsupportedOperationException();
    }

    /**
     * Answers an array with the same elements as the given collection, with the same
     * {@linkplain Collection#size() size} as the collection, and a component type of
     * the given class.
     * <p/>
     * If {@code collection} maintains an ordering for its elements, the order of
     * those elements will be preserved in the result.
     *
     * @param <T> the component type of the array to be created
     * @param <E> the type of the elements in the collection to convert
     * @param collection the collection to make an array from
     * @param componentType the desired type of the result array's components
     * @return an array of {@code collection}'s elements
     * @throws NullPointerException if either {@code collection} or {@code componentType}
     * is {@code null}
     * @throws IllegalArgumentException if {@code componentType} is {@link Void#TYPE}
     * @see Collection#toArray()
     * @see <a href="http://cleveralias.blogs.com/thought_spearmints/2005/02/creating_an_arr.html">Creating
     * an array from a generic Java collection</a>
     */
    public static <T, E extends T> T[] toArray( Collection<E> collection,
        Class<T> componentType ) {

        if ( componentType.isPrimitive() )
            throw new IllegalArgumentException( "primitive component type" );

        final T[] collectionAsArray =
            cast( Array.newInstance( componentType, collection.size() ) );

        collection.inject( 0, new BinaryFunctor<Integer, E, Integer>() {
            public Integer evaluate( Integer first, E second ) {
                collectionAsArray[first] = second;
                return first + 1;
            }
        } );

        return collectionAsArray;
    }

    /**
     * Answers a dictionary whose keys are the elements of the given collection and whose
     * values are the results of applying the given mapper to the associated element
     * from the collection.
     * <p/>
     * {@code elements} are traversed in the order specified by {@link
     * Collection#forEachDo(UnaryFunctor)}.
     * <p/>
     * This method would be on {@link Collection} if it weren't
     * <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6404691">for</a>
     * <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6218229">these</a>
     * <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6292765">bugs</a>.
     *
     * @param <E> the type of elements in the given collection
     * @param <V> the types of the values in the resultant dictionary
     * @param elements a collection
     * @param mapper a function to apply to each element of {@code elements}
     * @return a mapping from {@code elements} to mapped values
     * @throws NullPointerException if either {@code elements} or {@code mapper} is
     * {@code null}
     */
    public static <E, V> Dictionary<E, V> toDictionary( Collection<E> elements,
        final UnaryFunctor<? super E, ? extends V> mapper ) {

        ensureNotNull( elements, COLLECTION );
        ensureNotNull( mapper, MAPPER );

        final Dictionary<E, V> mappings = emptyDictionary();

        elements.forEachDo( new UnaryFunctor<E, Void>() {
            public Void evaluate( E argument ) {
                mappings.putAt( argument, mapper.evaluate( argument ) );
                return null;
            }
        } );

        return mappings;
    }

    /**
     * Answers a dictionary whose keys are the elements of the given collection and whose
     * values are the results of applying the given mapper to the associated element
     * from the collection.
     * <p/>
     * {@code elements} are traversed in the order specified by {@link
     * Collection#forEachDo(UnaryFunctor)}.
     * <p/>
     * This method would be on {@link Collection} if it weren't
     * <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6404691">for</a>
     * <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6218229">these</a>
     * <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6292765">bugs</a>.
     *
     * @param <E> the type of elements in the given collection
     * @param <V> the types of the values in the resultant dictionary
     * @param elements a collection
     * @param mapper a function to apply to each element of {@code elements}
     * @return a mapping from {@code elements} to mapped values
     * @throws NullPointerException if either {@code elements} or {@code mapper} is
     * {@code null}
     */
    public static <E, V> IdentityDictionary<E, V> toIdentityDictionary(
        Collection<E> elements, final UnaryFunctor<? super E, ? extends V> mapper ) {

        ensureNotNull( elements, COLLECTION );
        ensureNotNull( mapper, MAPPER );

        final IdentityDictionary<E, V> mappings = emptyIdentityDictionary();

        elements.forEachDo( new UnaryFunctor<E, Void>() {
            public Void evaluate( E argument ) {
                mappings.putAt( argument, mapper.evaluate( argument ) );
                return null;
            }
        } );

        return mappings;
    }
}
