/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import static jaggregate.Bag.*;
import static jaggregate.Set.emptySet;
import static jaggregate.internal.EquivalenceTester.*;

/**
 * Represents an unordered collection whose <dfn>elements</dfn> can be accessed using an
 * explicitly assigned external <dfn>key</dfn>.  <dfn>Key equivalence</dfn> is defined by
 * {@link Object#equals(Object) equals}.
 *
 * @param <K> a restriction on the types of the keys that may be contained in the
 * dictionary
 * @param <V> a restriction on the types of the values that may be contained in the
 * dictionary
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: Dictionary.java,v 1.5 2008/05/09 03:12:56 pholser Exp $
 */
public class Dictionary<K, V> extends AbstractDictionaryImpl<K, V>
    implements Serializable {

    private static final long serialVersionUID = -1L;

    /**
     * Creates an empty dictionary.
     */
    public Dictionary() {
        super( OBJECT_ATTRIBUTES );
    }

    /**
     * Creates a dictionary that contains the given associations.
     *
     * @param associations the associations to add to the new dictionary
     * @throws NullPointerException if {@code associations} is {@code null}
     */
    public Dictionary( AbstractDictionary<? extends K, ? extends V> associations ) {
        super( associations, OBJECT_ATTRIBUTES );
    }

    /**
     * Creates a dictionary that contains the given associations.
     *
     * @param associations the associations to add to the new dictionary
     * @throws NullPointerException if {@code associations} is {@code null}, or if
     * any of the elements of {@code associations} is {@code null}
     */
    public Dictionary( Pair<? extends K, ? extends V>... associations ) {
        super( OBJECT_ATTRIBUTES );

        for ( Pair<? extends K, ? extends V> each : associations )
            putAt( each.key(), each.value() );
    }

    /**
     * Creates an empty dictionary.
     *
     * @param <T> the type of keys allowed in the new dictionary
     * @param <U> the type of values allowed in the new dictionary
     * @return a new empty dictionary
     */
    public static <T, U> Dictionary<T, U> emptyDictionary() {
        return new Dictionary<T, U>();
    }

    /**
     * Creates a dictionary that contains the given associations.
     *
     * @param <T> the type of keys allowed in the new dictionary
     * @param <U> the type of values allowed in the new dictionary
     * @param associations the associations to add to the new dictionary
     * @return the new dictionary
     * @throws NullPointerException if {@code associations} is {@code null}
     */
    public static <T, U> Dictionary<T, U> dictionaryFrom(
        AbstractDictionary<? extends T, ? extends U> associations ) {

        return new Dictionary<T, U>( associations );
    }

    /**
     * Creates a dictionary that contains the given associations.
     *
     * @param <T> the type of keys allowed in the new dictionary
     * @param <U> the type of values allowed in the new dictionary
     * @param associations the associations to add to the new dictionary
     * @return the new dictionary
     * @throws NullPointerException if {@code associations} is {@code null}, or if
     * any of the elements of {@code associations} is {@code null}
     */
    public static <T, U> Dictionary<T, U> dictionaryFrom(
        Pair<? extends T, ? extends U>[] associations ) {

        return new Dictionary<T, U>( associations );
    }

    /**
     * Creates a dictionary that contains the given associations.
     *
     * @param <T> the type of keys allowed in the new dictionary
     * @param <U> the type of values allowed in the new dictionary
     * @param firstAssociation an association to add to the new dictionary
     * @param restOfAssociations other associations to add to the new dictionary
     * @return the new dictionary
     * @throws NullPointerException if {@code firstAssociation} is {@code null},
     * {@code associations} is {@code null}, or if any of the elements of
     * {@code associations} is {@code null}
     */
    public static <T, U> Dictionary<T, U> dictionaryWith(
        Pair<? extends T, ? extends U> firstAssociation,
        Pair<? extends T, ? extends U>... restOfAssociations ) {

        Dictionary<T, U> dictionary = dictionaryFrom( restOfAssociations );
        dictionary.putAt( firstAssociation.key(), firstAssociation.value() );
        return dictionary;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public <R> Bag<R> collect(
        UnaryFunctor<? super Pair<K, V>, ? extends R> transformer ) {

        return (Bag<R>) super.collect( transformer );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public <R> Dictionary<K, R> collectValues(
        UnaryFunctor<? super V, ? extends R> transformer ) {

        return (Dictionary<K, R>) super.collectValues( transformer );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Dictionary<K, V> reject( UnaryCondition<? super Pair<K, V>> discriminator ) {
        return (Dictionary<K, V>) super.reject( discriminator );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Dictionary<K, V> rejectValues( UnaryCondition<? super V> discriminator ) {
        return (Dictionary<K, V>) super.rejectValues( discriminator );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Dictionary<K, V> select( UnaryCondition<? super Pair<K, V>> discriminator ) {
        return (Dictionary<K, V>) super.select( discriminator );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Dictionary<K, V> selectValues( UnaryCondition<? super V> discriminator ) {
        return (Dictionary<K, V>) super.selectValues( discriminator );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected Class<? extends AbstractDictionary> getImplementationClass() {
        return Dictionary.class;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected <T, U> Dictionary<T, U> newEmptyDictionary() {
        return emptyDictionary();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected <R> Bag<R> newEmptyExtensibleResultCollection() {
        return emptyBag();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected <T> Set<T> newEmptySet() {
        return emptySet();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected String asString( Object keyOrValue ) {
        return String.valueOf( keyOrValue );
    }

    private void writeObject( ObjectOutputStream outputStream )
        throws IOException {

        serializeTo( outputStream );
    }

    private void readObject( ObjectInputStream inputStream )
        throws IOException, ClassNotFoundException {

        deserializeFrom( inputStream, OBJECT_ATTRIBUTES );
    }
}
