/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

/**
 * Represents a variable-sized collection of objects whose elements are ordered based on
 * the natural ordering of its elements, as defined by the interface {@link Comparable}.
 * Elements may be added, removed, or inserted, and can be accessed using external
 * integer keys.
 *
 * @param <E> a restriction on the types of elements that may be included in the
 * collection
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: NaturallySortedCollection.java,v 1.2 2008/03/26 18:00:33 pholser Exp $
 */
public class NaturallySortedCollection<E extends Comparable<? super E>>
    extends SortedCollection<E> {

    private static final long serialVersionUID = -1L;

    /**
     * Creates an empty naturally sorted collection.
     */
    public NaturallySortedCollection() {
        super( Comparing.<E>byNaturalOrdering() );
    }

    /**
     * Creates a naturally sorted collection containing the elements from the given
     * collection.
     *
     * @param newElements the elements to add to the new collection
     * @throws NullPointerException if {@code newElements} is {@code null}
     */
    public NaturallySortedCollection( Collection<? extends E> newElements ) {
        super( Comparing.<E>byNaturalOrdering(), newElements );
    }

    /**
     * Creates a naturally sorted collection containing the elements from the given
     * array.
     *
     * @param newElements the elements to add to the new collection
     * @throws NullPointerException if {@code newElements} is {@code null}
     */
    public NaturallySortedCollection( E... newElements ) {
        super( Comparing.<E>byNaturalOrdering(), newElements );
    }

    /**
     * Creates a naturally sorted collection containing the elements from the given
     * iterable.
     *
     * @param newElements the elements to add to the new collection
     * @throws NullPointerException if {@code newElements} is {@code null}
     */
    public NaturallySortedCollection( Iterable<? extends E> newElements ) {
        super( Comparing.<E>byNaturalOrdering(), newElements );
    }

    /**
     * Creates an empty naturally sorted collection.
     *
     * @param <T> the type of elements to add to the new sorted collection
     * @return a new sorted collection
     */
    public static <T extends Comparable<? super T>>
    NaturallySortedCollection<T> emptyNaturallySortedCollection() {
        return new NaturallySortedCollection<T>();
    }

    /**
     * Creates a naturally sorted collection containing the elements from the given
     * collection.
     *
     * @param <T> the type of elements to add to the new sorted collection
     * @param newElements the elements to add to the new collection
     * @return a new sorted collection
     * @throws NullPointerException if {@code newElements} is {@code null}
     */
    public static <T extends Comparable<? super T>>
    NaturallySortedCollection<T> naturallySortedCollectionFrom(
        Collection<? extends T> newElements ) {

        return new NaturallySortedCollection<T>( newElements );
    }

    /**
     * Creates a naturally sorted collection containing the elements from the given
     * array.
     *
     * @param <T> the type of elements to add to the new sorted collection
     * @param newElements the elements to add to the new collection
     * @return a new sorted collection
     * @throws NullPointerException if {@code newElements} is {@code null}
     */
    public static <T extends Comparable<? super T>>
    NaturallySortedCollection<T> naturallySortedCollectionWith( T... newElements ) {
        return new NaturallySortedCollection<T>( newElements );
    }

    /**
     * Creates a naturally sorted collection containing the elements from the given
     * iterable.
     *
     * @param <T> the type of elements to add to the new sorted collection
     * @param newElements the elements to add to the new collection
     * @return a new sorted collection
     * @throws NullPointerException if {@code newElements} is {@code null}
     */
    public static <T extends Comparable<? super T>>
    NaturallySortedCollection<T> naturallySortedCollectionFrom(
        Iterable<? extends T> newElements ) {

        return new NaturallySortedCollection<T>( newElements );
    }
}
