/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import static jaggregate.internal.ArgumentChecks.*;

/**
 * Utility class that offers useful operations on objects in general.
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: Objects.java,v 1.6 2008/10/03 19:01:23 pholser Exp $
 */
public class Objects {
    private static final String SAME_VALUE_TEMPLATE = "<%1$s>";

    /**
     * Discourages instantiation.
     *
     * @throws UnsupportedOperationException always
     */
    protected Objects() {
        throw new UnsupportedOperationException();
    }

    /**
     * Tells whether two object references are considered
     * {@linkplain #equals(Object) equivalent}.  Both references can be {@code null}.
     *
     * @param first first reference
     * @param second second reference
     * @return result of {@code first == null ? second == null : first.equals(second)}
     */
    public static boolean areEqual( Object first, Object second ) {
        return first == null ? second == null : first.equals( second );
    }

    /**
     * Answers a {@linkplain #hashCode() hash value} for the given target object.
     *
     * @param target the target object
     * @return {@code target}'s hash value, or {@code 0} if {@code target} is {@code
     * null}
     */
    public static int nullSafeHashCode( Object target ) {
        return target == null ? 0 : target.hashCode();
    }

    /**
     * Answers the {@linkplain #toString() string representation} for the target object,
     * as {@link Object}'s implementation would compute it, disregarding overrides.
     *
     * @param target the target object
     * @return the default string representation for {@code target}
     */
    public static String identityToString( Object target ) {
        if ( target == null )
            return String.valueOf( target );

        return new StringBuilder( target.getClass().getName() )
            .append( '@' )
            .append( Integer.toHexString( target.hashCode() ) )
            .toString();
    }

    /**
     * Answers a predicate which will answer {@code true} if its argument compares
     * {@linkplain Object#equals(Object) equivalent} to the given comparand.
     *
     * @param comparand the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code comparand}
     */
    public static UnaryPredicate<Object> equalTo( Object comparand ) {
        return new EqualTo().bindSecond( comparand );
    }

    /**
     * Answers a predicate which will answer {@code true} if its argument refers to the
     * same object as the given comparand.
     *
     * @param comparand the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code comparand}
     */
    public static UnaryPredicate<Object> sameAs( Object comparand ) {
        return new SameAs().bindSecond( comparand );
    }

    /**
     * Answers a predicate which will answer {@code true} if its argument is an
     * instance of the given comparand class.
     *
     * @param comparand the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code comparand}
     * @throws NullPointerException if {@code comparand} is {@code null}
     */
    public static UnaryPredicate<Object> isA( Class<?> comparand ) {
        ensureNotNull( comparand, "comparand class" );

        return new IsA().bindSecond( comparand );
    }

    /**
     * Answers a predicate which will answer {@code true} if its argument is a class in
     * the hierarchy of the given object's class.
     *
     * @param instance the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code instance}
     */
    public static UnaryPredicate<Class<?>> classifies( Object instance ) {
        return new IsA().bindFirst( instance );
    }

    private static class EqualTo extends BinaryPredicate<Object, Object> {
        public boolean matches( Object first, Object second ) {
            return areEqual( first, second );
        }

        @Override
        protected String getFirstBoundDescription() {
            return SAME_VALUE_TEMPLATE;
        }
    }

    private static class SameAs extends BinaryPredicate<Object, Object> {
        public boolean matches( Object first, Object second ) {
            return first == second;
        }

        @Override
        protected String getFirstBoundDescription() {
            return SAME_VALUE_TEMPLATE;
        }
    }

    private static class IsA extends BinaryPredicate<Object, Class<?>> {
        public boolean matches( Object first, Class<?> second ) {
            return second.isInstance( first );
        }

        @Override
        protected String getFirstBoundDescription() {
            return "a class of which <%1$s> is an instance";
        }

        @Override
        protected String getSecondBoundDescription() {
            return "an instance of <%1$s>";
        }
    }
}
