/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import java.io.Serializable;

import static jaggregate.Objects.*;

/**
 * A simple 2-tuple, typically used to represent the key-value pairs of an
 * {@link AbstractDictionary}.
 *
 * @param <K> the type of the pair's key
 * @param <V> the type of the pair's value
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: Pair.java,v 1.2 2008/03/26 18:00:34 pholser Exp $
 */
public class Pair<K, V> implements Serializable {
    private static final long serialVersionUID = -1L;

    protected V value;
    private final K key;

    /**
     * Creates a pair.
     *
     * @param key the new pair's key
     * @param value the new pair's value
     */
    public Pair( K key, V value ) {
        this.key = key;
        this.value = value;
    }

    /**
     * Creates a new pair with the given key and value.  Unlike the
     * {@linkplain Pair#Pair(Object,Object) constructor}, this generic factory method
     * allows type inference to work its magic and can save the caller from having to
     * specify the actual types of the pair's elements twice.  For example:
     * <pre><code>
     *   Pair&lt;Integer, Integer&gt; square = new Pair&lt;Integer, Integer&gt;( 3, 9 );
     * </code></pre>
     * versus
     * <pre>
     *   Pair&lt;Integer, Integer&gt; square = pairOf( 3, 9 );
     * </pre>
     *
     * @param <T> the type of the key in the resultant pair
     * @param <U> the type of the value in the resultant pair
     * @param key a key
     * @param value a value
     * @return a pair containing {@code key} and {@code value}
     */
    public static <T, U> Pair<T, U> pair( T key, U value ) {
        return new Pair<T, U>( key, value );
    }

    /**
     * Gives the pair's key.
     *
     * @return the key
     */
    public K key() {
        return key;
    }

    /**
     * Gives the pair's value.
     *
     * @return the value
     */
    public V value() {
        return value;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals( Object that ) {
        if ( this == that )
            return true;

        if ( !( that instanceof Pair<?, ?> ) )
            return false;

        Pair<?, ?> other = (Pair<?, ?>) that;
        return areEqual( key, other.key() ) && areEqual( value, other.value() );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return nullSafeHashCode( key ) ^ nullSafeHashCode( value );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        StringBuilder buffer = new StringBuilder( String.valueOf( key ) );
        buffer.append( '=' ).append( value );

        return buffer.toString();
    }
}
