/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

/**
 * Provides protocol for reading an ordered collection of objects whose
 * <dfn>elements</dfn> can be accessed using external integer <dfn>keys</dfn>.  The keys
 * are between zero (0) and the {@linkplain #size() number of elements in the collection}
 * minus one (1), inclusive.
 *
 * @param <E> a restriction on the types of elements that may be included in the
 * sequence
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: ReadOnlySequence.java,v 1.6 2008/05/12 23:49:07 pholser Exp $
 */
public interface ReadOnlySequence<E> extends Collection<E> {
    /**
     * {@inheritDoc}
     *
     * @return a sequence of transformations
     */
    <R> ReadOnlySequence<R> collect( UnaryFunctor<? super E, ? extends R> transformer );

    /**
     * {@inheritDoc}
     *
     * @return a sequence of rejects
     */
    ReadOnlySequence<E> reject( UnaryCondition<? super E> discriminator );

    /**
     * {@inheritDoc}
     *
     * @return a sequence of selections
     */
    ReadOnlySequence<E> select( UnaryCondition<? super E> discriminator );

    /**
     * Answers a new sequence containing all of this sequence's elements in their
     * original order followed by all of the elements of the given sequence, in their
     * original order.  The {@linkplain #size() size} of the new sequence is equal to
     * the sum of the {@linkplain #size() size}s of this sequence and
     * {@code otherSequence}.
     * <p/>
     * Sequences that enforce an ordering on their elements are permitted to refine
     * this message to reorder the result.
     *
     * @param otherSequence the sequence to concatenate
     * @return the result of the concatenation
     * @throws NullPointerException if {@code otherSequence} is {@code null}
     */
    ReadOnlySequence<E> concat( ReadOnlySequence<? extends E> otherSequence );

    /**
     * Answers the object immediately following the first element which is equivalent
     * to the given target in this sequence.  An element immediately follows another if
     * its index is one greater than that of the other.  The order used to determine
     * which of this sequence's elements is the first to be equivalent to {@code target}
     * is the traversal order defined by {@link #forEachDo(UnaryFunctor) forEachDo} for
     * this sequence.
     *
     * @param target the object to lookup
     * @return the object following {@code target}
     * @throws java.util.NoSuchElementException if the first occurrence of
     * {@code target} is the {@linkplain #last() last} element of this sequence, or if
     * this sequence does not {@linkplain #includes(Object) include} the given target
     */
    E after( E target );

    /**
     * Answers the element at the given position in this sequence.
     *
     * @param index the index to lookup
     * @return the corresponding element
     * @throws IndexOutOfBoundsException if {@code index < 0}, or {@code index >=}
     * this sequence's {@linkplain #size() size}
     */
    E at( int index );

    /**
     * Answers the object immediately preceding the first element which is equivalent to
     * the given target in this sequence.  An element immediately precedes another if
     * its index is one less than that of the other.
     *
     * @param target the object to lookup
     * @return the object preceding {@code target}
     * @throws java.util.NoSuchElementException if {@code target} is the {@linkplain
     * #first() first} element of this sequence, or if this sequence does not
     * {@linkplain #includes(Object) include} the given target
     */
    E before( E target );

    /**
     * Answers a new sequence containing the specified range of elements of this
     * sequence in their original order.  The element at index {@code start} in this
     * sequence is at index {@code 0} in the new sequence; the element at
     * {@code start + 1} is at index {@code 1}, etc.  If {@code stop < start}, then the
     * new sequence is {@linkplain #isEmpty() empty}.  Otherwise, the {@linkplain
     * #size() size} of the new sequence is {@link Math#max(int, int)
     * max(0, stop - start)}.
     * <p/>
     * The given indices must be non-negative.
     *
     * @param start the beginning index
     * @param stop the ending index
     * @return a new sequence containing the specified range of elements
     * @throws IndexOutOfBoundsException if {@code stop > start} and ({@code start < 0}
     * or {@code start >=} this sequence's {@linkplain #size() size}), or if
     * {@code stop > start} and ({@code stop < 0} or {@code stop >=} this sequence's
     * {@linkplain #size() size})
     */
    ReadOnlySequence<E> copyRange( int start, int stop );

    /**
     * Answers a new sequence with {@linkplain #size() size} one greater than the size
     * of this sequence, containing the elements of this sequence and a given new
     * element placed at the end.
     * <p/>
     * Unless specifically refined, this message is defined to answer an instance of the
     * same class as this sequence.
     * <p/>
     * Sequences that enforce an ordering on their elements are permitted to refine
     * this message to reorder the result.
     *
     * @param newElement element to add to the copy
     * @return copy of this sequence with {@code newElement} appended
     */
    ReadOnlySequence<E> copyWith( E newElement );

    /**
     * Answers a new sequence with all of the elements of this sequence that are not
     * equivalent to the given element, in their original order.
     *
     * @param oldElement the element to strip from the copy of this sequence
     * @return copy of this sequence with {@code oldElement} removed
     */
    ReadOnlySequence<E> copyWithout( E oldElement );

    /**
     * {@inheritDoc}
     * <p/>
     * {@code operation} is evaluated with each element of this sequence in indexed
     * order starting at {@code 0}.  The first element is at index {@code 0}, the
     * second at index {@code 1}, etc.  The index of the last element is equal to this
     * sequence's {@linkplain #size() size} minus 1.
     *
     * @throws NullPointerException if {@code operation} is {@code null}
     */
    void forEachDo( UnaryFunctor<? super E, ?> operation );

    /**
     * Answers the index of the first element which causes the given discriminator to
     * answer {@code true} when the element is used as a parameter.  Answers {@code -1}
     * if no such element is found.
     * <p/>
     * The elements are traversed in the order specified by {@link
     * #forEachDo(UnaryFunctor) forEachDo} for this sequence.
     *
     * @param discriminator the discriminator to evaluate
     * @return the index of the first match, or {@code -1} if there is no match
     * @throws NullPointerException if {@code discriminator} is {@code null}
     */
    int findFirst( UnaryCondition<? super E> discriminator );

    /**
     * Answers the index of the last element which causes the given discriminator to
     * answer {@code true} when the element is used as a parameter.  Answers {@code -1}
     * if no such element is found.
     * <p/>
     * The elements are traversed in the order specified by {@link
     * #forEachInReverseDo(UnaryFunctor) forEachInReverseDo} for this sequence.
     *
     * @param discriminator the discriminator to evaluate
     * @return the index of the last match, or {@code -1} if there is no match
     * @throws NullPointerException if {@code discriminator} is {@code null}
     */
    int findLast( UnaryCondition<? super E> discriminator );

    /**
     * Evaluates the given operation with each element of this sequence in the reverse
     * of this sequence's standard traversal order.  The elements are traversed in the
     * opposite order from {@link #forEachDo(UnaryFunctor) forEachDo}.  Each element
     * is visited exactly once.
     *
     * @param operation the operation to evaluate
     * @throws NullPointerException if {@code operation} is {@code null}
     */
    void forEachInReverseDo( UnaryFunctor<? super E, ?> operation );

    /**
     * Answers the element at index {@code 0} in this sequence.
     *
     * @return the first element in the sequence
     * @throws java.util.NoSuchElementException if this sequence is
     * {@linkplain #isEmpty() empty}
     */
    E first();

    /**
     * For those elements of this sequence between two given positions, inclusive,
     * evaluates the given operation with the element at that index as its argument.
     *
     * @param start beginning index
     * @param stop ending index
     * @param operation the operation to evaluate
     * @throws NullPointerException if {@code operation} is {@code null}
     * @throws IndexOutOfBoundsException if {@code start < 0} or {@code stop >=} this
     * sequence's {@linkplain #size() size}
     */
    void forEachInRangeDo( int start, int stop, UnaryFunctor<? super E, ?> operation );

    /**
     * For those elements of this sequence between two given positions, inclusive,
     * evaluates the given operation with an element of this sequence as the first
     * argument and the element's position (index) as the second.
     *
     * @param start beginning index
     * @param stop ending index
     * @param operation the operation to evaluate
     * @throws NullPointerException if {@code operation} is {@code null}
     * @throws IndexOutOfBoundsException if {@code start < 0} or {@code stop >=} this
     * sequence's {@linkplain #size() size}
     */
    void forEachInRangeDoWithIndex( int start, int stop,
        BinaryFunctor<? super Integer, ? super E, ?> operation );

    /**
     * Answers the index of the first element which is equivalent to the given target}.
     * Answers {@code -1} if no such element is found.
     * <p/>
     * The elements are traversed in the order specified by {@link
     * #forEachDo(UnaryFunctor) forEachDo} for this sequence.
     *
     * @param target element to match
     * @return index of first match, or {@code -1} if no match
     */
    int indexOf( E target );

    /**
     * Answers the index of the first element of this sequence which is the start
     * of a subsequence which matches a given target sequence.
     * <p/>
     * Each subsequence of this sequence starting at index {@code start} is checked for
     * a match with {@code targetSequence}.  To match, each element of a subsequence of
     * this sequence must be equivalent to the corresponding element of\
     * {@code targetSequence}.  Answers the index of the first element which begins a
     * matching subsequence; no further subsequences are considered.
     * Answers {@code -1} if no such subsequence is found in this sequence, or if
     * {@code targetSequence} is {@linkplain #isEmpty() empty}.
     * <p/>
     * The elements are traversed in the order specified by {@link
     * #forEachDo(UnaryFunctor) forEachDo} message for this sequence.
     *
     * @param targetSequence sequence to match
     * @param start index at which to begin the search
     * @return index of the beginning of the first match, or {@code -1} if none
     * @throws NullPointerException if {@code targetSequence} is {@code null}
     * @throws IndexOutOfBoundsException if {@code start < 0} or {@code start >=}
     * this sequence's {@linkplain #size() size}
     */
    int indexOf( ReadOnlySequence<? extends E> targetSequence, int start );

    /**
     * Evaluates the given operation with the index of each element of this sequence, in
     * order, together with the element itself.
     * <p/>
     * {@code operation} is evaluated with the index of each element of this sequence as
     * the first argument and the element itself as the second argument.  Evaluation is
     * in indexed order starting at {@code 0}.  The first element is at index {@code 0},
     * the second at index {@code 1}, etc.  The index of the last element is equal to
     * this sequence's {@linkplain #size() size} minus 1.
     *
     * @param operation the operation to perform
     * @throws NullPointerException if {@code operation} is {@code null}
     */
    void doWithIndex( BinaryFunctor<? super Integer, ? super E, ?> operation );

    /**
     * Answers the last element of this sequence, the element at the index equal to this
     * sequence's {@linkplain #size() size} minus 1.
     *
     * @return the last element in the sequence
     * @throws java.util.NoSuchElementException if this sequence is
     * {@linkplain #isEmpty() empty}
     */
    E last();

    /**
     * Answers a sequence with the elements of this sequence arranged in reverse order.
     * <p/>
     * This operation is equivalent to:
     * <ol>
     * <li>Create a new sequence which conforms to the same protocols as this sequence.
     * <li>Traverse the elements of this sequence in the order specified by
     * {@link #forEachInReverseDo(UnaryFunctor) forEachInReverseDo} message,
     * adding each element of this sequence to the new sequence.
     * <li>Answer the new sequence.
     * </ol>
     *
     * @return a new sequence with the elements of this sequence in reverse order
     */
    ReadOnlySequence<E> reverse();

    /**
     * For each element of this sequence and the corresponding element of the given
     * sequence, evaluates the given operation with this sequence's element as the first
     * parameter, and the element of the other sequence as the second parameter.  This
     * sequence and {@code otherSequence} must have the same {@linkplain #size() size}.
     * <p/>
     * The elements of this sequence and {@code otherSequence} are traversed in indexed
     * order starting at {@code 0}.  The {@code operation} is evaluated with the
     * elements at index {@code 0} in the two sequences, then index {@code 1}, etc.
     *
     * @param <T> a constraint on the types of elements in the given sequence
     * @param otherSequence the other elements to use in the evaluation
     * @param operation the operation to perform
     * @throws NullPointerException if either {@code otherSequence} or {@code operation}
     * is {@code null}
     * @throws IllegalArgumentException if this sequence's {@linkplain #size() size}
     * is not equal to the {@linkplain #size() size} of the other sequence
     */
    <T> void doWithOthers( ReadOnlySequence<? extends T> otherSequence,
        BinaryFunctor<? super E, ? super T, ?> operation );
}
