/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import java.util.NoSuchElementException;

import static jaggregate.OrderedCollection.*;
import static jaggregate.internal.ArgumentChecks.*;

/**
 * Represents a stream that has a positionable sequence of values that can be read. The
 * sequence values are provided by a sequenced collection that serves as the
 * <dfn>stream backing store</dfn>.
 * <p/>
 * Note that there is nothing to stop a caller to continue to manipulate the sequence
 * being streamed over via another reference that allows such manipulations: adding to
 * the sequence, replacing elements in the sequence, and so forth.  However, the stream
 * will still stop reading elements when it reaches the end of the sequence as it was
 * set up at the time the stream was created.
 *
 * @param <E> a restriction on the types of elements being streamed over
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: ReadStream.java,v 1.5 2008/10/03 19:01:23 pholser Exp $
 */
public class ReadStream<E> extends CollectionStream<E, ReadOnlySequence<E>>
    implements GettableStream<E, ReadOnlySequence<E>> {

    /**
     * Creates a new stream over the given backing sequence.
     *
     * @param backingStore the backing sequence to stream over
     * @throws NullPointerException if {@code backingStore} is {@code null}
     */
    public ReadStream( ReadOnlySequence<E> backingStore ) {
        super( backingStore );
    }

    /**
     * Creates a new stream over the given backing sequence.
     *
     * @param <T> the type of elements in the backing sequence
     * @param backingStore the backing sequence to stream over
     * @return the new stream
     * @throws NullPointerException if {@code backingStore} is {@code null}
     */
    public static <T> ReadStream<T> readStreamOver( ReadOnlySequence<T> backingStore ) {
        return new ReadStream<T>( backingStore );
    }

    /**
     * {@inheritDoc}
     */
    public <R> void forEachDo( UnaryFunctor<? super E, ? extends R> operation ) {
        ensureOpen();

        while ( !atEnd() )
            operation.evaluate( next() );
    }

    /**
     * {@inheritDoc}
     */
    public E next() {
        ensureOpen();

        if ( atEnd() )
            throw new NoSuchElementException( "no more elements in the stream" );

        E result = contents().at( position() );
        advance();

        return result;
    }

    /**
     * {@inheritDoc}
     */
    public OrderedCollection<E> next( int amount ) {
        ensureOpen();

        if ( amount < 0 )
            throw new IllegalArgumentException( NEGATIVE_AMOUNT + amount );

        int numberOfNextElements = maxPosition() - position();

        if ( amount > numberOfNextElements )
            throw new IllegalArgumentException(
                "there are not "
                + amount
                + " elements left in the stream, only "
                + numberOfNextElements );

        OrderedCollection<E> nextElements = emptyOrderedCollection();
        for ( int i = 0; i < amount; ++i )
            nextElements.add( next() );

        return nextElements;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isNextMatchFor( E target ) {
        ensureOpen();

        return Objects.areEqual( target, next() );
    }

    /**
     * {@inheritDoc}
     */
    public E peek() {
        ensureOpen();

        return atEnd() ? null : contents().at( position() );
    }

    /**
     * {@inheritDoc}
     */
    public boolean isPeekMatchFor( E target ) {
        ensureOpen();

        return Objects.areEqual( target, peek() );
    }

    /**
     * {@inheritDoc}
     */
    public void skip( int amount ) {
        ensureOpen();

        if ( amount < 0 )
            throw new IllegalArgumentException( NEGATIVE_AMOUNT + amount );

        int newPosition = position() + amount;
        if ( newPosition >= maxPosition() )
            setToEnd();
        else
            position( newPosition );
    }

    /**
     * {@inheritDoc}
     */
    public boolean skipTo( E target ) {
        ensureOpen();

        while ( !atEnd() ) {
            if ( Objects.areEqual( target, next() ) )
                return true;
        }

        return false;
    }

    /**
     * {@inheritDoc}
     */
    public OrderedCollection<E> upTo( E target ) {
        ensureOpen();

        OrderedCollection<E> results = emptyOrderedCollection();

        while ( !atEnd() ) {
            E nextItem = next();

            if ( Objects.areEqual( target, nextItem ) )
                return results;

            results.add( nextItem );
        }

        return results;
    }
}
