/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import static jaggregate.Dictionary.*;

/**
 * A set for which <dfn>equivalence</dfn> is defined by
 * {@link Object#equals(Object) equals}.
 *
 * @param <E> a restriction on the types of elements that may be included in the set
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: Set.java,v 1.4 2008/05/07 06:00:49 pholser Exp $
 */
public class Set<E> extends AbstractSet<E> implements Serializable {
    private static final long serialVersionUID = -1L;

    /**
     * Creates an empty set.
     */
    public Set() {
        // empty on purpose
    }

    /**
     * Creates a set containing the elements in the given collection.
     *
     * @param elements elements to add to the new set
     * @throws NullPointerException if {@code elements} is {@code null}
     */
    public Set( Collection<? extends E> elements ) {
        super( elements );
    }

    /**
     * Creates a set containing the elements in the given array.
     *
     * @param elements elements to add to the new set
     * @throws NullPointerException if {@code elements} is {@code null}
     */
    public Set( E... elements ) {
        super( elements );
    }

    /**
     * Creates a set containing the elements in the given iterable object.
     *
     * @param elements elements to add to the new set
     * @throws NullPointerException if {@code elements} is {@code null}
     */
    public Set( Iterable<? extends E> elements ) {
        super( elements );
    }

    /**
     * Creates an empty set.
     *
     * @param <T> the type of elements allowed in the new set
     * @return a new empty set
     */
    public static <T> Set<T> emptySet() {
        return new Set<T>();
    }

    /**
     * Creates a set containing the elements in the given collection.
     *
     * @param <T> the type of elements allowed in the new set
     * @param elements elements to add to the new set
     * @return the new set
     * @throws NullPointerException if {@code elements} is {@code null}
     */
    public static <T> Set<T> setFrom( Collection<? extends T> elements ) {
        return new Set<T>( elements );
    }

    /**
     * Creates a set containing the elements in the given array.
     *
     * @param <T> the type of elements allowed in the new set
     * @param elements elements to add to the new set
     * @return the new set
     * @throws NullPointerException if {@code elements} is {@code null}
     */
    public static <T> Set<T> setFrom( T[] elements ) {
        return new Set<T>( elements );
    }

    /**
     * Creates a set containing the given elements.
     *
     * @param <T> the type of elements allowed in the new set
     * @param newElement first new element to add
     * @param restOfNewElements remainder of the elements to add
     * @return the new set
     * @throws NullPointerException if {@code restOfNewElements} is {@code null}
     * @throws IllegalArgumentException if any of the new elements is found to violate
     * restrictions on the characteristics of valid elements
     * @see #addAll(Collection)
     */
    public static <T> Set<T> setWith( T newElement, T... restOfNewElements ) {
        Set<T> newSet = setFrom( restOfNewElements );
        newSet.add( newElement );

        return newSet;
    }

    /**
     * Creates a set containing the elements given by an iterable object.
     *
     * @param <T> the type of elements allowed in the new set
     * @param elements elements to add to the new set
     * @return the new set
     * @throws NullPointerException if {@code elements} is {@code null}
     */
    public static <T> Set<T> setFrom( Iterable<? extends T> elements ) {
        return new Set<T>( elements );
    }

    /**
     * {@inheritDoc}
     * <p/>
     * Answers self.
     */
    @Override
    public Set<E> toSet() {
        return this;
    }

    /**
     * {@inheritDoc}
     *
     * @return a set of results
     */
    @Override
    public <R> Set<R> collect( UnaryFunctor<? super E, ? extends R> transformer ) {
        return (Set<R>) super.collect( transformer );
    }

    /**
     * {@inheritDoc}
     *
     * @return a set of rejects
     */
    @Override
    public Set<E> reject( UnaryCondition<? super E> discriminator ) {
        return (Set<E>) super.reject( discriminator );
    }

    /**
     * {@inheritDoc}
     *
     * @return a set of selections
     */
    @Override
    public Set<E> select( UnaryCondition<? super E> discriminator ) {
        return (Set<E>) super.select( discriminator );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected Class<? extends AbstractSet> getImplementationClass() {
        return Set.class;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected AbstractDictionary<E, Object> newEmptyStorage() {
        return emptyDictionary();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected Set<E> newEmptyExtensibleCollection() {
        return newEmptyExtensibleResultCollection();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected <R> Set<R> newEmptyExtensibleResultCollection() {
        return emptySet();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected <T> Set<T> newEmptySet() {
        return emptySet();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected String asString( E anElement ) {
        return String.valueOf( anElement );
    }

    private void writeObject( ObjectOutputStream output ) throws IOException {
        serializeTo( output );
    }

    private void readObject( ObjectInputStream input )
        throws IOException, ClassNotFoundException {

        deserializeFrom( input );
    }
}
