/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

/**
 * Provides protocol for a stream consisting of a finite number of <dfn>past</dfn> and
 * <dfn>future sequence values</dfn>.  It maintains a position on its sequence values
 * and allows the position to be altered.
 *
 * @param <E> the type of elements in the stream
 * @param <S> the type of the backing store
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: Stream.java,v 1.2 2008/03/26 18:00:34 pholser Exp $
 */
public interface Stream<E, S extends ReadOnlySequence<E>> {
    /**
     * Tells whether this stream is at the end of its values.
     *
     * @return {@code true} if this stream has no future sequence values available for
     * reading; {@code false} otherwise.
     * @throws IllegalStateException if this stream has been previously
     * {@linkplain #close() closed}
     */
    boolean atEnd();

    /**
     * Disassociates this stream from its backing store.  Any system resources
     * associated with the association should be released.
     * <p/>
     * This message is <dfn>idempotent</dfn>: sending it multiple times to this stream
     * has no effect.
     */
    void close();

    /**
     * Answers a sequence containing the complete contents of this stream.
     * <p/>
     * The returned sequence contains this stream's past and future sequence values,
     * in order.  The {@link Collection#size() size} of the collection is the sum of
     * the sizes of the past and future sequence values.
     *
     * @return the contents of the receiving stream
     * @throws IllegalStateException if this stream has been previously
     * {@linkplain #close() closed}
     */
    S contents();

    /**
     * Tells whether there are any sequence values in this stream.
     *
     * @return {@code true} if there are any sequence values in this stream
     * @throws IllegalStateException if this stream has been previously
     * {@linkplain #close() closed}
     */
    boolean isEmpty();

    /**
     * Answers the current position of this stream.
     *
     * @return current position
     * @throws IllegalStateException if this stream has been previously
     * {@linkplain #close() closed}
     */
    int position();

    /**
     * Sets the current position in this stream.
     * <p/>
     * If the number of past sequence values is smaller than {@code amount},
     * move objects in sequence from the front of the future sequence values to the back
     * of the past sequence values until the number of past sequence values is equal to
     * {@code amount}.
     * <p/>
     * If the number of past sequence values is greater than {@code amount},
     * move objects in sequence from the back of the past sequence values to the front
     * of the future sequence values until the number past sequence values is equal to
     * {@code amount}.
     * <p/>
     * If the number of past sequence values is equal to {@code amount}, no action is
     * taken.
     *
     * @param amount the new position
     * @throws IndexOutOfBoundsException if {@code amount} is negative, or if this
     * stream is not {@link #isEmpty() empty} and {@code amount >=} the total number of
     * sequence values in this stream
     * @throws IllegalStateException if this stream has been previously
     * {@linkplain #close() closed}
     */
    void position( int amount );

    /**
     * Sets the position of this stream to be at the beginning of the stream of values.
     * <p/>
     * Sets the future sequence values to be the current past sequence values appended
     * with the current future sequence values--that is, makes this stream's past
     * sequence values empty.
     *
     * @throws IllegalStateException if this stream has been previously
     * {@linkplain #close() closed}
     */
    void reset();

    /**
     * Sets the position of this stream to be at the end of the stream of values.
     * <p/>
     * Adds all of the future sequence values are appended, in sequence, to the past
     * sequence values--that is, makes the future sequence values empty.
     *
     * @throws IllegalStateException if this stream has been previously
     * {@linkplain #close() closed}
     */
    void setToEnd();
}
