/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import java.util.regex.Pattern;

import static jaggregate.internal.ArgumentChecks.*;

/**
 * Utility class that offers common operations on strings.
 *
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: Strings.java,v 1.7 2008/10/03 19:01:23 pholser Exp $
 */
public class Strings {
    /**
     * A function which gives the length of its argument.
     *
     * @see String#length()
     */
    public static final UnaryFunctor<String, Integer> LENGTH =
        new UnaryFunctor<String, Integer>() {
            public Integer evaluate( String argument ) {
                return argument.length();
            }
        };

    /**
     * Discourages instantiation.
     *
     * @throws UnsupportedOperationException always
     */
    protected Strings() {
        throw new UnsupportedOperationException();
    }

    /**
     * Answers a predicate which will answer {@code true} if the given substring occurs
     * anywhere within its argument.
     *
     * @see String#contains(CharSequence)
     * @param substring the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code substring}
     * @throws NullPointerException if {@code substring} is {@code null}
     */
    public static UnaryPredicate<String> hasSubstring( CharSequence substring ) {
        ensureNotNull( substring, "substring comparand" );

        return new HasSubstring().bindSecond( substring );
    }

    /**
     * Answers a predicate which will answer {@code true} if the given string has the
     * predicate's argument as a substring.
     *
     * @see String#contains(CharSequence)
     * @param target the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code target}
     * @throws NullPointerException if {@code target} is {@code null}
     */
    public static UnaryPredicate<CharSequence> isSubstringOf( String target ) {
        ensureNotNull( target, "string comparand" );

        return new HasSubstring().bindFirst( target );
    }

    /**
     * Answers a predicate which will answer {@code true} if its argument has the same
     * string contents as the given comparand.
     *
     * @see String#contentEquals(CharSequence)
     * @param comparand the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code comparand}
     * @throws NullPointerException if {@code comparand} is {@code null}
     */
    public static UnaryPredicate<String> hasSameContentsAs( CharSequence comparand ) {
        ensureNotNull( comparand, "CharSequence comparand" );

        return new HasSameContents().bindSecond( comparand );
    }

    /**
     * Answers a predicate which will answer {@code true} if its argument ends with the
     * given suffix.
     *
     * @see String#endsWith(String)
     * @param suffix the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code suffix}
     * @throws NullPointerException if {@code suffix} is {@code null}
     */
    public static UnaryPredicate<String> endsWith( String suffix ) {
        ensureNotNull( suffix, "suffix" );

        return new EndsWith().bindSecond( suffix );
    }

    /**
     * Answers a predicate which will answer {@code true} if its argument is a suffix of
     * the given string.
     *
     * @see String#endsWith(String)
     * @param target the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code target}
     * @throws NullPointerException if {@code target} is {@code null}
     */
    public static UnaryPredicate<String> isSuffixOf( String target ) {
        ensureNotNull( target, "suffix target" );

        return new EndsWith().bindFirst( target );
    }

    /**
     * Answers a predicate which will answer {@code true} if its argument starts with
     * the given prefix.
     *
     * @param prefix the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code prefix}
     * @throws NullPointerException if {@code prefix} is {@code null}
     */
    public static UnaryPredicate<String> startsWith( String prefix ) {
        ensureNotNull( prefix, "prefix" );

        return new StartsWith().bindSecond( prefix );
    }

    /**
     * Answers a predicate which will answer {@code true} if its argument is a prefix
     * of the given target.
     *
     * @param target the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code target}
     * @throws NullPointerException if {@code target} is {@code null}
     */
    public static UnaryPredicate<String> isPrefixOf( String target ) {
        ensureNotNull( target, "prefix target" );

        return new StartsWith().bindFirst( target );
    }

    /**
     * Answers a predicate which will answer {@code true} if its argument is equivalent
     * to the given comparand, disregarding case.
     *
     * @see String#equalsIgnoreCase(String)
     * @param comparand the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code comparand}
     * @throws NullPointerException if {@code comparand} is {@code null}
     */
    public static UnaryPredicate<String> equalsIgnoreCase( String comparand ) {
        ensureNotNull( comparand, "comparand" );

        return new EqualsIgnoreCase().bindSecond( comparand );
    }

    /**
     * Answers a predicate which will answer {@code true} if any part of its argument
     * matches the given regular expression pattern.
     *
     * @see java.util.regex.Matcher#find()
     * @param pattern the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code pattern}
     * @throws NullPointerException if {@code pattern} is {@code null}
     */
    public static UnaryPredicate<CharSequence> matchesPattern( String pattern ) {
        ensureNotNull( pattern, "regex pattern" );

        return new MatchesPattern().bindSecond( pattern );
    }

    /**
     * Answers a predicate which will answer {@code true} if its regular expression
     * pattern argument matches the given target.
     *
     * @see java.util.regex.Matcher#find()
     * @param target the comparand to use with the returned predicate
     * @return a predicate that can be used to compare arguments with {@code target}
     * @throws NullPointerException if {@code target} is {@code null}
     */
    public static UnaryPredicate<String> patternMatches( CharSequence target ) {
        ensureNotNull( target, "regex pattern target" );

        return new MatchesPattern().bindFirst( target );
    }

    /**
     * Answers a function which will give the character value of a string target at the
     * given index.
     *
     * @see String#charAt(int)
     * @param index index into a target string
     * @return a function that will interrogate strings for the character at
     * {@code index}
     */
    public static UnaryFunction<String, Character> charAt( int index ) {
        return new CharAt().bindSecond( index );
    }

    /**
     * Answers a function which will give the first index of a given character value
     * in a given string.
     *
     * @see String#indexOf(int)
     * @param ch desired character in the target string
     * @return a function that will interrogate strings for the first index of
     * {@code ch}
     */
    public static UnaryFunction<String, Integer> indexOf( int ch ) {
        return new IndexOfCharacter().bindSecond( ch );
    }

    /**
     * Answers a function which will give the first index of a given substring
     * in a given string.
     *
     * @see String#indexOf(String)
     * @param substring desired substring in the target string
     * @return a function that will interrogate strings for the first index of
     * {@code substring}
     */
    public static UnaryFunction<String, Integer> indexOf( String substring ) {
        return new IndexOfString().bindSecond( substring );
    }

    /**
     * Answers a function which will give the last index of a given character value
     * in a given string.
     *
     * @see String#lastIndexOf(int)
     * @param ch desired character in the target string
     * @return a function that will interrogate strings for the last index of {@code ch}
     */
    public static UnaryFunction<String, Integer> lastIndexOf( int ch ) {
        return new LastIndexOfCharacter().bindSecond( ch );
    }

    /**
     * Answers a function which will give the last index of a given substring
     * in a given string.
     *
     * @see String#lastIndexOf(String)
     * @param substring desired substring in the target string
     * @return a function that will interrogate strings for the last index of
     * {@code substring}
     */
    public static UnaryFunction<String, Integer> lastIndexOf( String substring ) {
        return new LastIndexOfString().bindSecond( substring );
    }

    /**
     * Answers a function which will give the substring of a given string beginning at
     * the given index.
     *
     * @see String#substring(int)
     * @param index index at which to start the substring
     * @return a function that will interrogate strings for substrings
     */
    public static UnaryFunction<String, String> substring( int index ) {
        return new Substring().bindSecond( index );
    }

    private static class HasSubstring extends BinaryPredicate<String, CharSequence> {
        public boolean matches( String first, CharSequence second ) {
            return first.contains( second );
        }

        @Override
        protected String getFirstBoundDescription() {
            return "a string that is a substring of <%1$s>";
        }

        @Override
        protected String getSecondBoundDescription() {
            return "a string that contains the string <%1$s>";
        }
    }

    private static class HasSameContents extends BinaryPredicate<String, CharSequence> {
        public boolean matches( String first, CharSequence second ) {
            return first.contentEquals( second );
        }

        @Override
        protected String getFirstBoundDescription() {
            return "<%1$s>";
        }
    }

    private static class EndsWith extends BinaryPredicate<String, String> {
        public boolean matches( String first, String second ) {
            return first.endsWith( second );
        }

        @Override
        protected String getFirstBoundDescription() {
            return "a string that is a suffix of <%1$s>";
        }

        @Override
        protected String getSecondBoundDescription() {
            return "a string that ends with <%1$s>";
        }
    }

    private static class StartsWith extends BinaryPredicate<String, String> {
        public boolean matches( String first, String second ) {
            return first.startsWith( second );
        }

        @Override
        protected String getFirstBoundDescription() {
            return "a string that is a prefix of <%1$s>";
        }

        @Override
        protected String getSecondBoundDescription() {
            return "a string that starts with <%1$s>";
        }
    }

    private static class EqualsIgnoreCase extends BinaryPredicate<String, String> {
        public boolean matches( String first, String second ) {
            return first.equalsIgnoreCase( second );
        }

        @Override
        protected String getFirstBoundDescription() {
            return "a string equal to <%1$s> ignoring case";
        }
    }

    private static class MatchesPattern extends BinaryPredicate<CharSequence, String> {
        public boolean matches( CharSequence first, String second ) {
            return Pattern.compile( second ).matcher( first ).find();
        }

        @Override
        protected String getFirstBoundDescription() {
            return "a regex that matches <%1$s>";
        }

        @Override
        protected String getSecondBoundDescription() {
            return "a string that matches <%1$s>";
        }
    }

    private static class CharAt extends BinaryFunction<String, Integer, Character> {
        public Character evaluate( String first, Integer second ) {
            if ( second < 0 )
                return first.charAt( first.length() + second );

            return first.charAt( second );
        }
    }

    private static class IndexOfCharacter
        extends BinaryFunction<String, Integer, Integer> {

        public Integer evaluate( String first, Integer second ) {
            return first.indexOf( second );
        }
    }

    private static class IndexOfString extends BinaryFunction<String, String, Integer> {
        public Integer evaluate( String first, String second ) {
            return first.indexOf( second );
        }
    }

    private static class LastIndexOfCharacter
        extends BinaryFunction<String, Integer, Integer> {

        public Integer evaluate( String first, Integer second ) {
            return first.lastIndexOf( second );
        }
    }

    private static class LastIndexOfString
        extends BinaryFunction<String, String, Integer> {

        public Integer evaluate( String first, String second ) {
            return first.lastIndexOf( second );
        }
    }

    private static class Substring extends BinaryFunction<String, Integer, String> {
        public String evaluate( String first, Integer second ) {
            return first.substring( second );
        }
    }
}
