/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import static jaggregate.internal.ArgumentChecks.*;

/**
 * A functor that accepts one argument.
 *
 * @param <A> a constraint on the allowable types for the functor's argument
 * @param <R> a constraint on the allowable types for the functor's return value
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: UnaryFunction.java,v 1.6 2008/10/03 19:01:23 pholser Exp $
 */
public abstract class UnaryFunction<A, R> implements UnaryFunctor<A, R> {
    /**
     * Creates a new functor.
     */
    protected UnaryFunction() {
        // For subclasses.
    }

    /**
     * Answers a unary functor that evaluates in the same manner as this functor
     * would if handed the result of evaluating a given unary functor with a given
     * argument as its own argument; effectively offering a <dfn>composition</dfn>
     * of the two functors.
     *
     * @param <T> a constraint on the allowable types of the argument to the composed
     * functor
     * @param other the functor to compose
     * @return a composed unary functor
     * @throws NullPointerException if {@code other} is {@code null}
     */
    public final <T> UnaryFunction<T, R> compose( UnaryFunction<? super T, A> other ) {
        return new ComposedUnaryFunction<A, R, T>( this, other );
    }

    private static class ComposedUnaryFunction<A, R, T> extends UnaryFunction<T, R> {
        private final UnaryFunction<A, R> wrapper;
        private final UnaryFunction<? super T, A> composed;

        ComposedUnaryFunction( UnaryFunction<A, R> wrapper,
            UnaryFunction<? super T, A> composed ) {

            ensureNotNull( composed, "'right-hand' operand" );

            this.wrapper = wrapper;
            this.composed = composed;
        }

        public R evaluate( T argument ) {
            return wrapper.evaluate( composed.evaluate( argument ) );
        }
    }
}
