/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate;

import static java.lang.String.*;

import static jaggregate.internal.ArgumentChecks.*;

/**
 * A predicate that accepts one argument.
 *
 * @param <A> a constraint on the allowable types for the predicate's argument.
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: UnaryPredicate.java,v 1.6 2008/10/03 19:01:23 pholser Exp $
 */
public abstract class UnaryPredicate<A> implements UnaryCondition<A> {
    /**
     * Creates a new predicate.
     */
    protected UnaryPredicate() {
        // For subclasses.
    }

    /**
     * Answers a predicate that represents the logical <dfn>inverse</dfn> of this
     * predicate; wherever this predicate's {@link #matches(Object) matches} method
     * would answer {@code true}, the inverse answers {@code false}; and vice versa.
     *
     * @return the inverse of this predicate
     */
    public final UnaryPredicate<A> not() {
        return new InvertedUnaryPredicate<A>( this );
    }

    /**
     * Answers a predicate that represents the logical <dfn>inverse</dfn> of the given
     * predicate; wherever the given predicate's {@link #matches(Object) matches} method
     * would answer {@code true}, the inverse answers {@code false}; and vice versa.
     *
     * @param <T> constraint on the types accepted by the predicate
     * @param predicate the predicate to invert
     * @return the inverse of {@code predicate}
     * @throws NullPointerException if {@code predicate} is {@code null}
     */
    public static <T> UnaryPredicate<T> not( UnaryPredicate<T> predicate ) {
        return new InvertedUnaryPredicate<T>( predicate );
    }

    /**
     * Answers a predicate that represents the logical <dfn>disjunction</dfn> of this
     * predicate and another predicate.  The disjunction's {@link #matches(Object)
     * matches} method answers {@code true} if either this predicate or the other
     * predicate would answer {@code true}.
     * <p/>
     * When the disjunction is evaluated, this predicate is always evaluated first.
     * The other predicate may not be evaluated; so do not depend on its evaluation for
     * side effects.
     *
     * @param other the "right-hand" operand of the disjunction
     * @return the logical disjunction of this predicate and {@code other}
     * @throws NullPointerException if {@code other} is {@code null}
     */
    public final UnaryPredicate<A> or( UnaryPredicate<? super A> other ) {
        return new DisjunctiveUnaryPredicate<A>( this, other );
    }

    /**
     * Answers a predicate that represents the logical <dfn>conjunction</dfn> of this
     * predicate and another predicate.  The conjunction's {@link #matches(Object)
     * matches} method answers {@code true} if both this predicate and the other
     * predicate would answer {@code true}.
     * <p/>
     * When the conjunction is evaluated, this predicate is always evaluated first.
     * The other predicate may not be evaluated; so do not depend on its evaluation for
     * side effects.
     *
     * @param other the "right-hand" operand of the conjunction
     * @return the logical conjunction of this predicate and {@code other}
     * @throws NullPointerException if {@code other} is {@code null}
     */
    public final UnaryPredicate<A> and( UnaryPredicate<? super A> other ) {
        return new ConjunctiveUnaryPredicate<A>( this, other );
    }

    /**
     * Answers a predicate that represents the logical <dfn>exclusive disjunction</dfn>
     * of this predicate and another predicate.  The disjunction's
     * {@link #matches(Object) matches} method answers {@code true} if either this
     * predicate or the other predicate, but not both, would answer {@code true}.
     * <p/>
     * When the disjunction is evaluated, both this predicate and the other predicate
     * will be evaluated.
     *
     * @param other the "right-hand" operand of the disjunction
     * @return the logical exclusive disjunction of this predicate and {@code other}
     * @throws NullPointerException if {@code other} is {@code null}
     */
    public final UnaryPredicate<A> xor( UnaryPredicate<? super A> other ) {
        return new ExclusivelyDisjunctiveUnaryPredicate<A>( this, other );
    }

    /**
     * Gives a string that describes the kinds of objects which match this predicate.
     * <p/>
     * If not overridden, this method answers the same as {@link #toString()}.
     *
     * @return a description of objects that meet this predicate's criteria
     */
    public String describe() {
        return toString();
    }

    private static class InvertedUnaryPredicate<A> extends UnaryPredicate<A> {
        private final UnaryPredicate<A> wrapped;

        InvertedUnaryPredicate( UnaryPredicate<A> wrapped ) {
            ensureNotNull( wrapped, DISCRIMINATOR );
            this.wrapped = wrapped;
        }

        public boolean matches( A target ) {
            return !wrapped.matches( target );
        }

        @Override
        public String describe() {
            return format( "something other than <%1$s>", wrapped.describe() );
        }
    }

    private abstract static class CompositeUnaryPredicate<A> extends UnaryPredicate<A> {
        protected final UnaryPredicate<A> first;
        protected final UnaryPredicate<? super A> second;

        protected CompositeUnaryPredicate( UnaryPredicate<A> first,
            UnaryPredicate<? super A> second ) {

            ensureNotNull( second, "'right-hand' operand" );

            this.first = first;
            this.second = second;
        }
    }

    private static class DisjunctiveUnaryPredicate<A>
        extends CompositeUnaryPredicate<A> {

        DisjunctiveUnaryPredicate( UnaryPredicate<A> first,
            UnaryPredicate<? super A> second ) {

            super( first, second );
        }

        public boolean matches( A target ) {
            return first.matches( target ) || second.matches( target );
        }

        @Override
        public String describe() {
            return format( "<%1$s> or <%2$s>", first.describe(), second.describe() );
        }
    }

    private static class ConjunctiveUnaryPredicate<A>
        extends CompositeUnaryPredicate<A> {

        ConjunctiveUnaryPredicate( UnaryPredicate<A> first,
            UnaryPredicate<? super A> second ) {

            super( first, second );
        }

        public boolean matches( A target ) {
            return first.matches( target ) && second.matches( target );
        }

        @Override
        public String describe() {
            return format( "<%1$s> and <%2$s>", first.describe(), second.describe() );
        }
    }

    private static class ExclusivelyDisjunctiveUnaryPredicate<A>
        extends CompositeUnaryPredicate<A> {

        ExclusivelyDisjunctiveUnaryPredicate( UnaryPredicate<A> first,
            UnaryPredicate<? super A> second ) {

            super( first, second );
        }

        public boolean matches( A target ) {
            return first.matches( target ) ^ second.matches( target );
        }

        @Override
        public String describe() {
            return format( "either <%1$s> or <%2$s> but not both", first.describe(),
                second.describe() );
        }
    }
}
