/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate.internal;

import jaggregate.BinaryFunctor;
import jaggregate.OrderedCollection;
import jaggregate.ReadOnlySequence;
import jaggregate.UnaryCondition;
import jaggregate.UnaryFunctor;
import static jaggregate.Objects.*;
import static jaggregate.OrderedCollection.*;
import static jaggregate.internal.ArgumentChecks.*;

/**
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: ReadOnlySequences.java,v 1.8 2008/10/03 19:01:23 pholser Exp $
 */
public class ReadOnlySequences {
    /**
     * Discourages instantiation.
     *
     * @throws UnsupportedOperationException always
     */
    protected ReadOnlySequences() {
        throw new UnsupportedOperationException();
    }

    /**
     * Two {@link jaggregate.ReadOnlySequence}s are considered equivalent if they have
     * the same {@linkplain jaggregate.Collection#size() size} and if the corresponding
     * elements at each index are {@linkplain Object#equals(Object) equivalent}.
     *
     * @see Object#equals(Object)
     * @param first first alleged sequence
     * @param second second alleged sequence
     * @return whether {@code first} and {@code second} are equivalent
     */
    public static boolean areSequencesEqual( Object first, Object second ) {
        if ( first == second )
            return true;

        if ( !( first instanceof ReadOnlySequence<?> )
            || !( second instanceof ReadOnlySequence<?> ) )
            return false;

        final ReadOnlySequence<?> firstSequence = (ReadOnlySequence<?>) first;
        final ReadOnlySequence<?> secondSequence = (ReadOnlySequence<?>) second;

        boolean sequencesMatch = false;
        if ( firstSequence.size() == secondSequence.size() ) {
            try {
                firstSequence.doWithIndex(
                    new SequencesEqualAtEachIndex( firstSequence, secondSequence ) );

                sequencesMatch = true;
            }
            catch ( NonLocalReturnException ignored ) {
                sequencesMatch = false;
            }
        }

        return sequencesMatch;
    }

    /**
     * Computes a hash value for a given sequence.
     *
     * @see Object#hashCode()
     * @param sequence the sequence to hash
     * @return the computed hash value for {@code sequence}
     */
    public static int hashCode( ReadOnlySequence<?> sequence ) {
        if ( sequence == null )
            return 0;

        return sequence.inject( 0, new BinaryFunctor<Integer, Object, Integer>() {
            public Integer evaluate( Integer first, Object second ) {
                return first + nullSafeHashCode( second );
            }
        } );
    }

    /**
     * Computes a string representation for a given sequence.
     *
     * @see Object#toString()
     * @param sequence the sequence to stringify
     * @return the stringification of {@code sequence}
     */
    public static String toString( final ReadOnlySequence<?> sequence ) {
        if ( sequence == null )
            return "null";

        final StringBuilder buffer = new StringBuilder( "[" );

        sequence.doWithIndex( new BinaryFunctor<Integer, Object, Void>() {
            public Void evaluate( Integer first, Object second ) {
                buffer.append( second );

                if ( first < sequence.size() - 1 )
                    buffer.append( ',' );

                return null;
            }
        } );

        buffer.append( ']' );

        return buffer.toString();
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#concat(jaggregate.ReadOnlySequence)}
     */
    public static <E> OrderedCollection<E> concat( ReadOnlySequence<? extends E> first,
        ReadOnlySequence<? extends E> second ) {

        ensureNotNull( first, SEQUENCE );
        ensureNotNull( second, SEQUENCE );

        OrderedCollection<E> concatenation = orderedCollectionFrom( first );
        concatenation.addAll( second );

        return concatenation;
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#copyRange(int, int)}
     */
    public static <E> OrderedCollection<E> copyRange( ReadOnlySequence<E> sequence,
        int start, int stop ) {

        ensureNotNull( sequence, SEQUENCE );

        if ( start > stop )
            return emptyOrderedCollection();

        if ( stop > start ) {
            checkIndexIsValidOn( sequence, start, STARTING_INDEX );
            checkIndexIsValidOn( sequence, stop, ENDING_INDEX );
        }

        final OrderedCollection<E> subsequence = emptyOrderedCollection();

        forEachInRangeDo( sequence, start, stop, new UnaryFunctor<E, Void>() {
            public Void evaluate( E nextElement ) {
                subsequence.add( nextElement );
                return null;
            }
        } );

        return subsequence;
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#copyWith(Object)}
     */
    public static <E> OrderedCollection<E> copyWith( ReadOnlySequence<E> sequence,
        E newElement ) {

        ensureNotNull( sequence, SEQUENCE );

        OrderedCollection<E> copy = orderedCollectionFrom( sequence );
        copy.add( newElement );

        return copy;
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#copyWithout(Object)}
     */
    public static <E> OrderedCollection<E> copyWithout( ReadOnlySequence<E> sequence,
        final E oldElement ) {

        ensureNotNull( sequence, SEQUENCE );

        final OrderedCollection<E> copy = emptyOrderedCollection();

        sequence.forEachDo( new UnaryFunctor<E, Void>() {
            public Void evaluate( E nextElement ) {
                if ( !areEqual( nextElement, oldElement ) )
                    copy.add( nextElement );
                return null;
            }
        } );

        return copy;
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#doWithIndex(jaggregate.BinaryFunctor)}
     */
    public static <E> void doWithIndex( ReadOnlySequence<E> sequence,
        BinaryFunctor<? super Integer, ? super E, ?> operation ) {

        ensureNotNull( sequence, SEQUENCE );
        ensureNotNull( operation, OPERATION );

        for ( int i = 0; i < sequence.size(); ++i )
            operation.evaluate( i, sequence.at( i ) );
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#doWithOthers(jaggregate.ReadOnlySequence,
     * jaggregate.BinaryFunctor)}
     */
    public static <E, T> void doWithOthers( ReadOnlySequence<E> first,
        final ReadOnlySequence<? extends T> second,
        final BinaryFunctor<? super E, ? super T, ?> operation ) {

        ensureNotNull( first, SEQUENCE );
        ensureNotNull( second, SEQUENCE );
        ensureNotNull( operation, OPERATION );

        if ( first.size() != second.size() ) {
            throw new IllegalArgumentException( EXPECTED_SEQUENCE_SIZE + first.size()
                + BUT_WAS + second.size() );
        }

        first.doWithIndex(
            new BinaryFunctor<Integer, E, Void>() {
                public Void evaluate( Integer index, E nextElement ) {
                    operation.evaluate( nextElement, second.at( index ) );
                    return null;
                }
            }
        );
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#forEachInRangeDo(int, int,
     * jaggregate.UnaryFunctor)}
     */
    public static <E> void forEachInRangeDo( ReadOnlySequence<E> sequence, int start,
        int stop, UnaryFunctor<? super E, ?> operation ) {

        ensureNotNull( sequence, SEQUENCE );
        ensureNotNull( operation, OPERATION );
        ensureIndexIsGreaterThan( start, -1, STARTING_INDEX );

        if ( stop != start )
            ensureIndexIsLessThan( stop, sequence.size(), ENDING_INDEX );

        for ( int i = start; i <= stop; ++i )
            operation.evaluate( sequence.at( i ) );
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#forEachInRangeDoWithIndex(int, int,
     * jaggregate.BinaryFunctor)}
     */
    public static <E> void forEachInRangeDoWithIndex( ReadOnlySequence<E> sequence,
        int start, int stop,
        BinaryFunctor<? super Integer, ? super E, ?> operation ) {

        ensureNotNull( sequence, SEQUENCE );
        ensureNotNull( operation, OPERATION );
        ensureIndexIsGreaterThan( start, -1, STARTING_INDEX );
        ensureIndexIsLessThan( stop, sequence.size(), ENDING_INDEX );

        for ( int i = start; i <= stop; ++i )
            operation.evaluate( i, sequence.at( i ) );
    }

    /**
     * @see {@link
     * jaggregate.ReadOnlySequence#forEachInReverseDo(jaggregate.UnaryFunctor)}
     */
    public static <E> void forEachInReverseDo( ReadOnlySequence<E> sequence,
        UnaryFunctor<? super E, ?> operation ) {

        ensureNotNull( sequence, SEQUENCE );
        ensureNotNull( operation, OPERATION );

        for ( int i = sequence.size() - 1; i >= 0; --i )
            operation.evaluate( sequence.at( i ) );
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#findFirst(jaggregate.UnaryCondition)}
     */
    public static <E> int findFirst( ReadOnlySequence<E> sequence,
        final UnaryCondition<? super E> discriminator ) {

        ensureNotNull( sequence, SEQUENCE );
        ensureNotNull( discriminator, DISCRIMINATOR );

        final int[] indexOfMatch = { -1 };

        try {
            sequence.doWithIndex(
                new BinaryFunctor<Integer, E, Void>() {
                    public Void evaluate( Integer first, E second ) {
                        if ( discriminator.matches( second ) ) {
                            indexOfMatch[ 0 ] = first;
                            throw new NonLocalReturnException();
                        }

                        return null;
                    }
                }
            );
        }
        catch ( NonLocalReturnException ignored ) {
            return indexOfMatch[ 0 ];
        }

        return -1;
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#findLast(jaggregate.UnaryCondition)}
     */
    public static <E> int findLast( ReadOnlySequence<E> sequence,
        final UnaryCondition<? super E> discriminator ) {

        ensureNotNull( sequence, SEQUENCE );
        ensureNotNull( discriminator, DISCRIMINATOR );

        final int[] indexOfMatch = { sequence.size() - 1 };

        try {
            sequence.forEachInReverseDo( new UnaryFunctor<E, Void>() {
                public Void evaluate( E argument ) {
                    if ( discriminator.matches( argument ) )
                        throw new NonLocalReturnException();

                    --indexOfMatch[ 0 ];
                    return null;
                }
            } );
        }
        catch ( NonLocalReturnException ignored ) {
            return indexOfMatch[ 0 ];
        }

        return -1;
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#indexOf(Object)}
     */
    public static <E> int indexOf( ReadOnlySequence<E> sequence, final E target ) {
        ensureNotNull( sequence, SEQUENCE );

        final int[] indexOfMatch = { -1 };

        try {
            sequence.doWithIndex(
                new BinaryFunctor<Integer, E, Void>() {
                    public Void evaluate( Integer first, E second ) {
                        if ( areEqual( target, second ) ) {
                            indexOfMatch[ 0 ] = first;
                            throw new NonLocalReturnException();
                        }

                        return null;
                    }
                } );

            return -1;
        }
        catch ( NonLocalReturnException ignored ) {
            return indexOfMatch[ 0 ];
        }
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#indexOf(jaggregate.ReadOnlySequence, int)}
     */
    public static <E> int indexOf( ReadOnlySequence<E> sequence,
        ReadOnlySequence<? extends E> target, int start ) {

        ensureNotNull( sequence, SEQUENCE );
        ensureNotNull( target, "target sequence" );
        checkIndexIsValidOn( sequence, start, STARTING_INDEX );

        if ( target.isEmpty() || target.size() + start > sequence.size() )
            return -1;

        for ( int startingIndex = start;
            startingIndex < sequence.size() - target.size() + 1;
            ++startingIndex ) {

            if ( areEqual( sequence.at( startingIndex ), target.first() ) ) {
                for ( int index = 0;
                    areEqual( sequence.at( startingIndex + index ), target.at( index ) );
                    ++index ) {

                    if ( index == target.size() - 1 )
                        return startingIndex;
                }
            }
        }

        return -1;
    }

    /**
     * @see {@link jaggregate.ReadOnlySequence#reverse()}
     */
    public static <E> OrderedCollection<E> reverse( ReadOnlySequence<E> sequence ) {
        ensureNotNull( sequence, SEQUENCE );

        final OrderedCollection<E> reversed = emptyOrderedCollection();

        sequence.forEachInReverseDo( new UnaryFunctor<E, Void>() {
            public Void evaluate( E argument ) {
                reversed.add( argument );
                return null;
            }
        } );

        return reversed;
    }

    /**
     * Raises an exception if the given index is found to be invalid on the given
     * sequence.
     *
     * @param sequence the sequence to address
     * @param index the index to check
     * @param failureMessage the message to add to the exception if {@code index} is
     * invalid on {@code sequence}
     * @throws SequenceIndexOutOfBoundsException if invalid index
     */
    public static void checkIndexIsValidOn( ReadOnlySequence<?> sequence, int index,
        String failureMessage ) {

        ensureIndexIsGreaterThan( index, -1, failureMessage );
        ensureIndexIsLessThan( index, sequence.size(), failureMessage );
    }

    /**
     * Raises an exception if the given index is found to be less than or equal to the
     * given boundary.
     *
     * @param index the index to check
     * @param boundary the boundary against which to check
     * @param failureMessage the message to add to the exception if {@code index} is
     * less than or equal to {@code boundary}
     * @throws SequenceIndexOutOfBoundsException if invalid index
     */
    public static void ensureIndexIsGreaterThan( int index, int boundary,
        String failureMessage ) {

        if ( index <= boundary )
            throw new SequenceIndexOutOfBoundsException( failureMessage, index );
    }

    /**
     * Raises an exception if the given index is found to be greater than or equal to the
     * given boundary.
     *
     * @param index the index to check
     * @param boundary the boundary against which to check
     * @param failureMessage the message to add to the exception if {@code index} is
     * greater than or equal to {@code boundary}
     * @throws SequenceIndexOutOfBoundsException if invalid index
     */
    public static void ensureIndexIsLessThan( int index, int boundary,
        String failureMessage ) {

        if ( index >= boundary )
            throw new SequenceIndexOutOfBoundsException( failureMessage, index );
    }

    private static class SequencesEqualAtEachIndex
        implements BinaryFunctor<Integer, Object, Void> {

        private final ReadOnlySequence<?> firstSequence;
        private final ReadOnlySequence<?> secondSequence;

        public SequencesEqualAtEachIndex( ReadOnlySequence<?> firstSequence,
            ReadOnlySequence<?> secondSequence ) {

            this.firstSequence = firstSequence;
            this.secondSequence = secondSequence;
        }

        public Void evaluate( Integer first, Object second ) {
            Object firstAtIndex = firstSequence.at( first );
            Object secondAtIndex = secondSequence.at( first );
            if ( !areEqual( firstAtIndex, secondAtIndex ) )
                throw new NonLocalReturnException();

            return null;
        }
    }
}
