/*
 Copyright 2004-2008 Paul R. Holser, Jr.  All rights reserved.
 Licensed under the Academic Free License version 3.0
 */

package jaggregate.internal.number;

import java.io.Serializable;

/**
 * Wrapper for basic Java numeric types, that provides a uniform interface with which
 * {@link jaggregate.Interval Interval}s can do arithmetic.
 *
 * @param <E> the type of {@link Number Number} that is wrapped
 * @author <a href="mailto:pholser@alumni.rice.edu">Paul Holser</a>
 * @version $Id: OperableNumber.java,v 1.1 2008/03/26 18:01:27 pholser Exp $
 */
public abstract class OperableNumber<E extends Number> implements Serializable {
    private static final long serialVersionUID = -1L;

    private E wrapped;

    /**
     * Creates a new number wrapper.
     *
     * @param wrapped the wrapped number
     */
    protected OperableNumber( E wrapped ) {
        this.wrapped = wrapped;
    }

    /**
     * Tells whether this number is equal to zero.
     *
     * @return {@code true} if this number is zero
     */
    public abstract boolean isZero();

    /**
     * Adds the given number to this one, returning the sum.
     *
     * @param addend the number to add to this one
     * @return the sum
     */
    public abstract OperableNumber<E> plus( E addend );

    /**
     * Adds the given integer to this number, returning the sum.
     *
     * @param addend the integer to add to this number
     * @return the sum
     */
    public abstract OperableNumber<E> plus( int addend );

    /**
     * Adds the given number to this one, returning the sum.
     *
     * @param addend the number to add to this one
     * @return the sum
     */
    public final OperableNumber<E> plus( OperableNumber<E> addend ) {
        return plus( addend.unwrapped() );
    }

    /**
     * Subtracts the given number from this one, returning the difference.
     *
     * @param subtrahend the number to subtract from this one
     * @return the difference
     */
    public abstract OperableNumber<E> minus( E subtrahend );

    /**
     * Subtracts the given number from this one, returning the difference.
     *
     * @param subtrahend the number to subtract from this one
     * @return the difference
     */
    public final OperableNumber<E> minus( OperableNumber<E> subtrahend ) {
        return minus( subtrahend.unwrapped() );
    }

    /**
     * Multiplies the given number with this one, returning the product.
     *
     * @param multiplier the number to multiply with this one
     * @return the product
     */
    public abstract OperableNumber<E> times( E multiplier );

    /**
     * Multiplies the given integer with this number, returning the product.
     *
     * @param multiplier the integer to multiply with this number
     * @return the product
     */
    public abstract OperableNumber<E> times( int multiplier );

    /**
     * Multiplies the given number with this one, returning the product.
     *
     * @param multiplier the number to multiply with this one
     * @return the product
     */
    public final OperableNumber<E> times( OperableNumber<E> multiplier ) {
        return times( multiplier.unwrapped() );
    }

    /**
     * Divides the given number into this one, returning the quotient.
     *
     * @param divisor the number to divide into this one
     * @return the quotient
     */
    public abstract OperableNumber<E> dividedBy( E divisor );

    /**
     * Divides the given number into this one, returning the quotient.
     *
     * @param divisor the number to divide into this one
     * @return the quotient
     */
    public final OperableNumber<E> dividedBy( OperableNumber<E> divisor ) {
        return dividedBy( divisor.unwrapped() );
    }

    /**
     * Divides the given number into this one, returning the modulus.
     *
     * @param divisor the number to divide into this one
     * @return the modulus
     */
    public abstract OperableNumber<E> modulo( E divisor );

    /**
     * Divides the given number into this one, returning the modulus.
     *
     * @param divisor the number to divide into this one
     * @return the modulus
     */
    public final OperableNumber<E> modulo( OperableNumber<E> divisor ) {
        return modulo( divisor.unwrapped() );
    }

    /**
     * Tells whether this number is positive.
     *
     * @return {@code true} if this number is positive
     */
    public abstract boolean isPositive();

    /**
     * Tells whether this number is negative.
     *
     * @return {@code true} if this number is negative
     */
    public abstract boolean isNegative();

    /**
     * Tells whether this number is greater than another.
     *
     * @param comparand the number to compare this one with
     * @return {@code true} if this number is {@code > comparand}
     */
    public abstract boolean greaterThan( E comparand );

    /**
     * Tells whether this number is greater than another.
     *
     * @param comparand the number to compare this one with
     * @return {@code true} if this number is {@code > comparand}
     */
    public final boolean greaterThan( OperableNumber<E> comparand ) {
        return greaterThan( comparand.unwrapped() );
    }

    /**
     * Tells whether this number is greater than or equal to another.
     *
     * @param comparand the number to compare this one with
     * @return {@code true} if this number is {@code >= comparand}
     */
    public abstract boolean greaterThanOrEqualTo( E comparand );

    /**
     * Tells whether this number is greater than or equal to another.
     *
     * @param comparand the number to compare this one with
     * @return {@code true} if this number is {@code >= comparand}
     */
    public final boolean greaterThanOrEqualTo( OperableNumber<E> comparand ) {
        return greaterThanOrEqualTo( comparand.unwrapped() );
    }

    /**
     * Tells whether this number is less than another.
     *
     * @param comparand the number to compare this one with
     * @return {@code true} if this number is {@code < comparand}
     */
    public abstract boolean lessThan( E comparand );

    /**
     * Tells whether this number is less than another.
     *
     * @param comparand the number to compare this one with
     * @return {@code true} if this number is {@code < comparand}
     */
    public final boolean lessThan( OperableNumber<E> comparand ) {
        return lessThan( comparand.unwrapped() );
    }

    /**
     * Tells whether this number is less than or equal to another.
     *
     * @param comparand the number to compare this one with
     * @return {@code true} if this number is {@code <= comparand}
     */
    public abstract boolean lessThanOrEqualTo( E comparand );

    /**
     * Tells whether this number is less than or equal to another.
     *
     * @param comparand the number to compare this one with
     * @return {@code true} if this number is {@code <= comparand}
     */
    public final boolean lessThanOrEqualTo( OperableNumber<E> comparand ) {
        return lessThanOrEqualTo( comparand.unwrapped() );
    }

    /**
     * Gives the number that is the multiplicative inverse of this one.
     *
     * @return this number, negated
     */
    public abstract OperableNumber<E> negated();

    /**
     * Increments this number by the given amount.
     *
     * @param increment the amount to increment this number by
     * @throws UnsupportedOperationException if the receiver is immutable.  The
     * instances returned by the "one" methods are immutable.
     * @see #one()
     */
    public abstract void incrementBy( OperableNumber<E> increment );

    /**
     * Gives an immutable number that is equal to one.
     *
     * @return an immutable, one-equivalent number
     */
    public abstract OperableNumber<E> one();

    /**
     * Gives the number that this one wraps.
     *
     * @return the wrapped number
     */
    public final E unwrapped() {
        return wrapped;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals( Object that ) {
        if ( this == that )
            return true;

        if ( that == null || !getClass().equals( that.getClass() ) )
            return false;

        OperableNumber<?> other = (OperableNumber<?>) that;
        return wrapped.equals( other.unwrapped() );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return wrapped.hashCode();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final String toString() {
        return wrapped.toString();
    }

    /**
     * Changes the numbers wrapped by this instance.
     *
     * @param newValue the new wrapped number
     */
    protected final void update( E newValue ) {
        wrapped = newValue;
    }
}
